import datetime
import logging
import time
from copy import deepcopy
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, TypeVar, Union

from pymilvus.client import entity_helper, utils
from pymilvus.client.abstract import LoopBase
from pymilvus.client.search_result import Hits
from pymilvus.exceptions import (
    MilvusException,
    ParamError,
)
from pymilvus.grpc_gen import milvus_pb2 as milvus_types

from .connections import Connections
from .constants import (
    BATCH_SIZE,
    CALC_DIST_BM25,
    CALC_DIST_COSINE,
    CALC_DIST_HAMMING,
    CALC_DIST_IP,
    CALC_DIST_JACCARD,
    CALC_DIST_L2,
    CALC_DIST_TANIMOTO,
    COLLECTION_ID,
    DEFAULT_SEARCH_EXTENSION_RATE,
    EF,
    FIELDS,
    GUARANTEE_TIMESTAMP,
    INT64_MAX,
    IS_PRIMARY,
    ITERATOR_FIELD,
    ITERATOR_SESSION_CP_FILE,
    ITERATOR_SESSION_TS_FIELD,
    MAX_BATCH_SIZE,
    MAX_FILTERED_IDS_COUNT_ITERATION,
    MAX_TRY_TIME,
    METRIC_TYPE,
    MILVUS_LIMIT,
    OFFSET,
    PARAMS,
    RADIUS,
    RANGE_FILTER,
    REDUCE_STOP_FOR_BEST,
    UNLIMITED,
)
from .schema import CollectionSchema
from .types import DataType
from .utility import mkts_from_datetime

log = logging.getLogger(__name__)
QueryIterator = TypeVar("QueryIterator")
SearchIterator = TypeVar("SearchIterator")


def fall_back_to_latest_session_ts():
    d = datetime.datetime.now()
    return mkts_from_datetime(d, milliseconds=1000.0)


def assert_info(condition: bool, message: str):
    if not condition:
        raise MilvusException(message)


def io_operation(io_func: Callable[[Any], None], message: str):
    try:
        io_func()
    except OSError as ose:
        raise MilvusException(message=message) from ose


def extend_batch_size(batch_size: int, next_param: dict, to_extend_batch_size: bool) -> int:
    extend_rate = 1
    if to_extend_batch_size:
        extend_rate = DEFAULT_SEARCH_EXTENSION_RATE
    if EF in next_param[PARAMS]:
        return min(MAX_BATCH_SIZE, batch_size * extend_rate, next_param[PARAMS][EF])
    return min(MAX_BATCH_SIZE, batch_size * extend_rate)


def check_set_flag(obj: Any, flag_name: str, kwargs: Dict[str, Any], key: str):
    setattr(obj, flag_name, kwargs.get(key, False))


class QueryIterator:
    def __init__(
        self,
        connection: Connections,
        collection_name: str,
        batch_size: Optional[int] = 1000,
        limit: Optional[int] = -1,
        expr: Optional[str] = None,
        output_fields: Optional[List[str]] = None,
        partition_names: Optional[List[str]] = None,
        schema: Optional[CollectionSchema] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> QueryIterator:
        self._conn = connection
        self._collection_name = collection_name
        self.__set_up_collection_id()
        self._output_fields = output_fields
        self._partition_names = partition_names
        self._schema = schema
        self._timeout = timeout
        self._session_ts = 0
        self._kwargs = kwargs
        self._kwargs[ITERATOR_FIELD] = "True"
        self._kwargs[COLLECTION_ID] = self._collection_id
        self.__check_set_batch_size(batch_size)
        self._limit = limit
        self.__check_set_reduce_stop_for_best()
        self._returned_count = 0
        self.__setup__pk_prop()
        self.__set_up_expr(expr)
        self._next_id = None
        self._cache_id_in_use = NO_CACHE_ID
        self._cp_file_handler = None
        self.__set_up_ts_cp()
        self.__seek_to_offset()

    def __set_up_collection_id(self):
        res = self._conn.describe_collection(self._collection_name)
        self._collection_id = res[COLLECTION_ID]

    def __seek_to_offset(self):
        # read pk cursor from cp file, no need to seek offset
        if self._next_id is not None:
            return
        offset = self._kwargs.get(OFFSET, 0)
        if offset > 0:
            seek_params = self._kwargs.copy()
            seek_params[OFFSET] = 0
            seek_params[ITERATOR_FIELD] = "False"
            seek_params[REDUCE_STOP_FOR_BEST] = "False"
            start_time = time.time()

            def seek_offset_by_batch(batch: int, expr: str) -> int:
                seek_params[MILVUS_LIMIT] = batch
                res = self._conn.query(
                    collection_name=self._collection_name,
                    expr=expr,
                    output_field=[],
                    partition_name=self._partition_names,
                    timeout=self._timeout,
                    **seek_params,
                )
                self.__update_cursor(res)
                return len(res)

            while offset > 0:
                batch_size = min(MAX_BATCH_SIZE, offset)
                next_expr = self.__setup_next_expr()
                seeked_count = seek_offset_by_batch(batch_size, next_expr)
                log.debug(
                    f"seeked offset, seek_expr:{next_expr} batch_size:{batch_size} seeked_count:{seeked_count}"
                )
                if seeked_count == 0:
                    log.info(
                        "seek offset has drained all matched results for query iterator, break"
                    )
                    break
                offset -= seeked_count
            self._kwargs[OFFSET] = 0
            seek_offset_duration = time.time() - start_time
            log.info(
                f"Finish seek offset for query iterator, offset:{offset}, current_pk_cursor:{self._next_id}, "
                f"duration:{seek_offset_duration}"
            )

    def __init_cp_file_handler(self) -> bool:
        mode = "w"
        if self._cp_file_path.exists():
            mode = "r+"
        try:
            self._cp_file_handler = self._cp_file_path.open(mode)
        except OSError as ose:
            raise MilvusException(
                message=f"Failed to open cp file for iterator:{self._cp_file_path_str}"
            ) from ose
        return mode == "r+"

    def __save_mvcc_ts(self):
        assert_info(
            self._cp_file_handler is not None,
            "Must init cp file handler before saving session_ts",
        )
        self._cp_file_handler.writelines(str(self._session_ts) + "\n")

    def __save_pk_cursor(self):
        if self._need_save_cp and self._next_id is not None:
            if not self._cp_file_path.exists():
                self._cp_file_handler.close()
                self._cp_file_handler = self._cp_file_path.open("w")
                self._buffer_cursor_lines_number = 0
                self.__save_mvcc_ts()
                log.warning(
                    "iterator cp file is not existed any more, recreate for iteration, "
                    "do not remove this file manually!"
                )
            if self._buffer_cursor_lines_number >= 100:
                self._cp_file_handler.seek(0)
                self._cp_file_handler.truncate()
                log.info(
                    "cursor lines in cp file has exceeded 100 lines, truncate the file and rewrite"
                )
                self._buffer_cursor_lines_number = 0
            self._cp_file_handler.writelines(str(self._next_id) + "\n")
            self._cp_file_handler.flush()
            self._buffer_cursor_lines_number += 1

    def __check_set_reduce_stop_for_best(self):
        if self._kwargs.get(REDUCE_STOP_FOR_BEST, True):
            self._kwargs[REDUCE_STOP_FOR_BEST] = "True"
        else:
            self._kwargs[REDUCE_STOP_FOR_BEST] = "False"

    def __check_set_batch_size(self, batch_size: int):
        if batch_size < 0:
            raise ParamError(message="batch size cannot be less than zero")
        if batch_size > MAX_BATCH_SIZE:
            raise ParamError(message=f"batch size cannot be larger than {MAX_BATCH_SIZE}")
        self._kwargs[BATCH_SIZE] = batch_size
        self._kwargs[MILVUS_LIMIT] = batch_size

    # rely on pk prop, so this method should be called after __setup__pk_prop
    def __set_up_expr(self, expr: str):
        if expr is not None:
            self._expr = expr
        elif self._pk_str:
            self._expr = self._pk_field_name + ' != ""'
        else:
            self._expr = self._pk_field_name + " < " + str(INT64_MAX)

    def __setup_ts_by_request(self):
        init_ts_kwargs = self._kwargs.copy()
        init_ts_kwargs[OFFSET] = 0
        init_ts_kwargs[MILVUS_LIMIT] = 1
        # just to set up mvccTs for iterator, no need correct limit
        res = self._conn.query(
            collection_name=self._collection_name,
            expr=self._expr,
            output_field=self._output_fields,
            partition_name=self._partition_names,
            timeout=self._timeout,
            **init_ts_kwargs,
        )
        if res is None:
            raise MilvusException(
                message="failed to connect to milvus for setting up "
                "mvccTs, check milvus servers' status"
            )
        if res.extra is not None:
            self._session_ts = res.extra.get(ITERATOR_SESSION_TS_FIELD, 0)
        if self._session_ts <= 0:
            log.warning("failed to get mvccTs from milvus server, use client-side ts instead")
            self._session_ts = fall_back_to_latest_session_ts()
        self._kwargs[GUARANTEE_TIMESTAMP] = self._session_ts

    def __set_up_ts_cp(self):
        self._buffer_cursor_lines_number = 0
        self._cp_file_path_str = self._kwargs.get(ITERATOR_SESSION_CP_FILE, None)
        self._cp_file_path = None
        # no input cp_file, set up mvccTs by query request
        if self._cp_file_path_str is None:
            self._need_save_cp = False
            self.__setup_ts_by_request()
        else:
            self._need_save_cp = True
            self._cp_file_path = Path(self._cp_file_path_str)
            if not self.__init_cp_file_handler():
                # input cp file is empty, set up mvccTs by query request
                self.__setup_ts_by_request()
                io_operation(self.__save_mvcc_ts, "Failed to save mvcc ts")
            else:
                try:
                    # input cp file is not emtpy, init mvccTs by reading cp file
                    lines = self._cp_file_handler.readlines()
                    line_count = len(lines)
                    if line_count < 2:
                        raise ParamError(
                            message=f"input cp file:{self._cp_file_path_str} should contain "
                            f"at least two lines, but only:{line_count} lines"
                        )
                    self._session_ts = int(lines[0])
                    self._kwargs[GUARANTEE_TIMESTAMP] = self._session_ts
                    if line_count > 1:
                        self._buffer_cursor_lines_number = line_count - 1
                        self._next_id = lines[self._buffer_cursor_lines_number].strip()
                except OSError as ose:
                    raise MilvusException(
                        message=f"Failed to read cp info from file:{self._cp_file_path_str}"
                    ) from ose
                except ValueError as e:
                    raise ParamError(message=f"cannot parse input cp session_ts:{lines[0]}") from e

    def __maybe_cache(self, result: List):
        if len(result) < 2 * self._kwargs[BATCH_SIZE]:
            return
        start = self._kwargs[BATCH_SIZE]
        cache_result = result[start:]
        cache_id = iterator_cache.cache(cache_result, NO_CACHE_ID)
        self._cache_id_in_use = cache_id

    def __is_res_sufficient(self, res: List):
        return res is not None and len(res) >= self._kwargs[BATCH_SIZE]

    def get_cursor(self) -> milvus_types.QueryCursor:
        cursor = milvus_types.QueryCursor
        cursor.session_ts = self._session_ts
        if self._pk_str:
            cursor.str_pk = str(self._next_id)
        else:
            cursor.int_pk = self._next_id
        return cursor

    def next(self):
        cached_res = iterator_cache.fetch_cache(self._cache_id_in_use)
        ret = None
        if self.__is_res_sufficient(cached_res):
            ret = cached_res[0 : self._kwargs[BATCH_SIZE]]
            res_to_cache = cached_res[self._kwargs[BATCH_SIZE] :]
            iterator_cache.cache(res_to_cache, self._cache_id_in_use)
        else:
            iterator_cache.release_cache(self._cache_id_in_use)
            current_expr = self.__setup_next_expr()
            log.debug(f"query_iterator_next_expr:{current_expr}")
            res = self._conn.query(
                collection_name=self._collection_name,
                expr=current_expr,
                output_fields=self._output_fields,
                partition_names=self._partition_names,
                timeout=self._timeout,
                **self._kwargs,
            )
            self.__maybe_cache(res)
            ret = res[0 : min(self._kwargs[BATCH_SIZE], len(res))]

        ret = self.__check_reached_limit(ret)
        self.__update_cursor(ret)
        io_operation(self.__save_pk_cursor, "failed to save pk cursor")
        self._returned_count += len(ret)
        return ret

    def __check_reached_limit(self, ret: List):
        if self._limit == UNLIMITED:
            return ret
        left_count = self._limit - self._returned_count
        if left_count >= len(ret):
            return ret
        # has exceeded the limit, cut off the result and return
        return ret[0:left_count]

    def __setup__pk_prop(self):
        fields = self._schema[FIELDS]
        for field in fields:
            if field.get(IS_PRIMARY):
                if field["type"] == DataType.VARCHAR:
                    self._pk_str = True
                else:
                    self._pk_str = False
                self._pk_field_name = field["name"]
                break
        if self._pk_field_name is None or self._pk_field_name == "":
            raise MilvusException(message="schema must contain pk field, broke")

    def __setup_next_expr(self) -> str:
        current_expr = self._expr
        if self._next_id is None:
            return current_expr
        filtered_pk_str = ""
        if self._pk_str:
            filtered_pk_str = f'{self._pk_field_name} > "{self._next_id}"'
        else:
            filtered_pk_str = f"{self._pk_field_name} > {self._next_id}"
        if current_expr is None or len(current_expr) == 0:
            return filtered_pk_str
        return "(" + current_expr + ")" + " and " + filtered_pk_str

    def __update_cursor(self, res: List) -> None:
        if len(res) == 0:
            return
        self._next_id = res[-1][self._pk_field_name]

    def close(self) -> None:
        # release cache in use
        iterator_cache.release_cache(self._cache_id_in_use)
        if self._cp_file_handler is not None:

            def inner_close():
                self._cp_file_handler.close()
                self._cp_file_path.unlink()
                log.info(f"removed cp file:{self._cp_file_path_str} for query iterator")

            io_operation(
                inner_close, f"failed to clear cp file:{self._cp_file_path_str} for query iterator"
            )


def metrics_positive_related(metrics: str) -> bool:
    if metrics in [CALC_DIST_L2, CALC_DIST_JACCARD, CALC_DIST_HAMMING, CALC_DIST_TANIMOTO]:
        return True
    if metrics in [CALC_DIST_IP, CALC_DIST_COSINE, CALC_DIST_BM25]:
        return False
    raise MilvusException(message=f"unsupported metrics type for search iteration: {metrics}")


class SearchPage(LoopBase):
    """Since we only support nq=1 in search iteration, so search iteration response
    should be different from raw response of search operation"""

    def __init__(self, res: Hits, session_ts: Optional[int] = 0):
        super().__init__()
        self._session_ts = session_ts
        self._results = []
        if res is not None:
            self._results.append(res)

    def get_session_ts(self):
        return self._session_ts

    def get_res(self):
        return self._results

    def __len__(self):
        length = 0
        for res in self._results:
            length += len(res)
        return length

    def get__item(self, idx: Any):
        if len(self._results) == 0:
            return None
        if idx >= self.__len__():
            msg = "Index out of range"
            raise IndexError(msg)
        index = 0
        ret = None
        for res in self._results:
            if index + len(res) <= idx:
                index += len(res)
            else:
                ret = res[idx - index]
                break
        return ret

    def merge(self, others: List[Hits]):
        if others is not None:
            for other in others:
                self._results.append(other)

    def ids(self):
        ids = []
        for res in self._results:
            for hit in res:
                ids.append(hit.id)
        return ids

    def distances(self):
        distances = []
        for res in self._results:
            for hit in res:
                distances.append(hit.distance)
        return distances


class SearchIterator:
    def __init__(
        self,
        connection: Connections,
        collection_name: str,
        data: Union[List, utils.SparseMatrixInputType],
        ann_field: str,
        param: Dict,
        batch_size: Optional[int] = 1000,
        limit: Optional[int] = UNLIMITED,
        expr: Optional[str] = None,
        partition_names: Optional[List[str]] = None,
        output_fields: Optional[List[str]] = None,
        timeout: Optional[float] = None,
        round_decimal: int = -1,
        schema: Optional[CollectionSchema] = None,
        **kwargs,
    ) -> SearchIterator:
        rows = entity_helper.get_input_num_rows(data)
        if rows > 1:
            raise ParamError(
                message="Not support search iteration over multiple vectors at present"
            )
        if rows == 0:
            raise ParamError(message="vector_data for search cannot be empty")
        self._conn = connection
        self._iterator_params = {
            "collection_name": collection_name,
            "data": data,
            "ann_field": ann_field,
            BATCH_SIZE: batch_size,
            "output_fields": output_fields,
            "partition_names": partition_names,
            "timeout": timeout,
            "round_decimal": round_decimal,
        }
        self._collection_name = collection_name
        self._expr = expr
        self.__check_set_params(param)
        self.__check_for_special_index_param()
        self._kwargs = kwargs
        self._kwargs[ITERATOR_FIELD] = "True"
        self.__set_up_collection_id()
        self._kwargs[COLLECTION_ID] = self._collection_id
        self._filtered_ids = []
        self._filtered_distance = None
        self._schema = schema
        self._limit = limit
        self._returned_count = 0
        self.__check_metrics()
        self.__check_offset()
        self.__check_rm_range_search_parameters()
        self.__setup__pk_prop()
        self.__init_search_iterator()

    def __set_up_collection_id(self):
        res = self._conn.describe_collection(self._collection_name)
        self._collection_id = res[COLLECTION_ID]

    def __init_search_iterator(self):
        init_page = self.__execute_next_search(self._param, self._expr, False)
        self._session_ts = init_page.get_session_ts()
        if self._session_ts <= 0:
            log.warning("failed to set up mvccTs from milvus server, use client-side ts instead")
            self._session_ts = fall_back_to_latest_session_ts()
        self._kwargs[GUARANTEE_TIMESTAMP] = self._session_ts
        if len(init_page) == 0:
            message = (
                "Cannot init search iterator because init page contains no matched rows, "
                "please check the radius and range_filter set up by searchParams"
            )
            log.error(message)
            self._cache_id = NO_CACHE_ID
            self._init_success = False
            return
        self._cache_id = iterator_cache.cache(init_page, NO_CACHE_ID)
        self.__set_up_range_parameters(init_page)
        self.__update_filtered_ids(init_page)
        self._init_success = True

    def __update_width(self, page: SearchPage):
        first_hit, last_hit = page[0], page[-1]
        if metrics_positive_related(self._param[METRIC_TYPE]):
            self._width = last_hit.distance - first_hit.distance
        else:
            self._width = first_hit.distance - last_hit.distance
        if self._width == 0.0:
            self._width = 0.05
            # enable a minimum value for width to avoid radius and range_filter equal error

    def __set_up_range_parameters(self, page: SearchPage):
        self.__update_width(page)
        self._tail_band = page[-1].distance
        log.debug(
            f"set up init parameter for searchIterator width:{self._width} tail_band:{self._tail_band}"
        )

    def __check_reached_limit(self) -> bool:
        if self._limit == UNLIMITED or self._returned_count < self._limit:
            return False
        log.debug(
            f"reached search limit:{self._limit}, returned_count:{self._returned_count}, directly return"
        )
        return True

    def __check_set_params(self, param: Dict):
        if param is None:
            self._param = {}
        else:
            self._param = deepcopy(param)
        if PARAMS not in self._param:
            self._param[PARAMS] = {}

    def __check_for_special_index_param(self):
        if (
            EF in self._param[PARAMS]
            and self._param[PARAMS][EF] < self._iterator_params[BATCH_SIZE]
        ):
            raise MilvusException(
                message="When using hnsw index, provided ef must be larger than or equal to batch size"
            )

    def __setup__pk_prop(self):
        fields = self._schema[FIELDS]
        for field in fields:
            if field.get(IS_PRIMARY):
                if field["type"] == DataType.VARCHAR:
                    self._pk_str = True
                else:
                    self._pk_str = False
                self._pk_field_name = field["name"]
                break
        if self._pk_field_name is None or self._pk_field_name == "":
            raise ParamError(message="schema must contain pk field, broke")

    def __check_metrics(self):
        if self._param[METRIC_TYPE] is None or self._param[METRIC_TYPE] == "":
            raise ParamError(message="must specify metrics type for search iterator")

    """we use search && range search to implement search iterator,
    so range search parameters are disabled to clients"""

    def __check_rm_range_search_parameters(self):
        if (
            (PARAMS in self._param)
            and (RADIUS in self._param[PARAMS])
            and (RANGE_FILTER in self._param[PARAMS])
        ):
            radius = self._param[PARAMS][RADIUS]
            range_filter = self._param[PARAMS][RANGE_FILTER]
            if metrics_positive_related(self._param[METRIC_TYPE]) and radius <= range_filter:
                raise MilvusException(
                    message=f"for metrics:{self._param[METRIC_TYPE]}, radius must be "
                    f"larger than range_filter, please adjust your parameter"
                )
            if not metrics_positive_related(self._param[METRIC_TYPE]) and radius >= range_filter:
                raise MilvusException(
                    message=f"for metrics:{self._param[METRIC_TYPE]}, radius must be "
                    f"smaller than range_filter, please adjust your parameter"
                )

    def __check_offset(self):
        if self._kwargs.get(OFFSET, 0) != 0:
            raise ParamError(message="Not support offset when searching iteration")

    def __update_filtered_ids(self, res: SearchPage):
        if len(res) == 0:
            return
        last_hit = res[-1]
        if last_hit is None:
            return
        if last_hit.distance != self._filtered_distance:
            self._filtered_ids = []  # distance has changed, clear filter_ids array
            self._filtered_distance = last_hit.distance  # renew the distance for filtering
        for hit in res:
            if hit.distance == last_hit.distance:
                self._filtered_ids.append(hit.id)
        if len(self._filtered_ids) > MAX_FILTERED_IDS_COUNT_ITERATION:
            raise MilvusException(
                message=f"filtered ids length has accumulated to more than "
                f"{MAX_FILTERED_IDS_COUNT_ITERATION!s}, "
                f"there is a danger of overly memory consumption"
            )

    def __is_cache_enough(self, count: int) -> bool:
        cached_page = iterator_cache.fetch_cache(self._cache_id)
        return cached_page is not None and len(cached_page) >= count

    def __extract_page_from_cache(self, count: int) -> SearchPage:
        cached_page = iterator_cache.fetch_cache(self._cache_id)
        if cached_page is None or len(cached_page) < count:
            raise ParamError(
                message=f"Wrong, try to extract {count} result from cache, "
                f"more than {len(cached_page)} there must be sth wrong with code"
            )

        ret_page_res = cached_page[0:count]
        ret_page = SearchPage(ret_page_res)
        left_cache_page = SearchPage(cached_page[count:])
        iterator_cache.cache(left_cache_page, self._cache_id)
        return ret_page

    def __push_new_page_to_cache(self, page: SearchPage) -> int:
        if page is None:
            raise ParamError(message="Cannot push None page into cache")
        cached_page: SearchPage = iterator_cache.fetch_cache(self._cache_id)
        if cached_page is None:
            iterator_cache.cache(page, self._cache_id)
            cached_page = page
        else:
            cached_page.merge(page.get_res())
        return len(cached_page)

    def next(self):
        # 0. check reached limit
        if not self._init_success or self.__check_reached_limit():
            return SearchPage(None)
        ret_len = self._iterator_params[BATCH_SIZE]
        if self._limit is not UNLIMITED:
            left_len = self._limit - self._returned_count
            ret_len = min(left_len, ret_len)

        # 1. if cached page is sufficient, directly return
        if self.__is_cache_enough(ret_len):
            ret_page = self.__extract_page_from_cache(ret_len)
            self._returned_count += len(ret_page)
            return ret_page

        # 2. if cached page not enough, try to fill the result by probing with constant width
        # until finish filling or exceeding max trial time: 10
        new_page = self.__try_search_fill()
        cached_page_len = self.__push_new_page_to_cache(new_page)
        ret_len = min(cached_page_len, ret_len)
        ret_page = self.__extract_page_from_cache(ret_len)
        if len(ret_page) == self._iterator_params[BATCH_SIZE]:
            self.__update_width(ret_page)

        # 3. update filter ids to avoid returning result repeatedly
        self._returned_count += ret_len
        return ret_page

    def __try_search_fill(self) -> SearchPage:
        final_page = SearchPage(None)
        try_time = 0
        coefficient = 1
        while True:
            next_params = self.__next_params(coefficient)
            next_expr = self.__filtered_duplicated_result_expr(self._expr)
            new_page = self.__execute_next_search(next_params, next_expr, True)
            self.__update_filtered_ids(new_page)
            try_time += 1
            if len(new_page) > 0:
                final_page.merge(new_page.get_res())
                self._tail_band = new_page[-1].distance
            if len(final_page) >= self._iterator_params[BATCH_SIZE]:
                break
            if try_time > MAX_TRY_TIME:
                log.warning(f"Search probe exceed max try times:{MAX_TRY_TIME} directly break")
                break
            # if there's a ring containing no vectors matched, then we need to extend
            # the ring continually to avoid empty ring problem
            coefficient += 1
        return final_page

    def __execute_next_search(
        self, next_params: dict, next_expr: str, to_extend_batch: bool
    ) -> SearchPage:
        log.debug(f"search_iterator_next_expr:{next_expr}, next_params:{next_params}")
        res = self._conn.search(
            self._iterator_params["collection_name"],
            self._iterator_params["data"],
            self._iterator_params["ann_field"],
            next_params,
            extend_batch_size(self._iterator_params[BATCH_SIZE], next_params, to_extend_batch),
            next_expr,
            self._iterator_params["partition_names"],
            self._iterator_params["output_fields"],
            self._iterator_params["round_decimal"],
            timeout=self._iterator_params["timeout"],
            schema=self._schema,
            **self._kwargs,
        )
        return SearchPage(res[0], res.get_session_ts())

    # at present, the range_filter parameter means 'larger/less and equal',
    # so there would be vectors with same distances returned multiple times in different pages
    # we need to refine and remove these results before returning
    def __filtered_duplicated_result_expr(self, expr: str):
        if len(self._filtered_ids) == 0:
            return expr

        filtered_ids_str = ""
        for filtered_id in self._filtered_ids:
            if self._pk_str:
                filtered_ids_str += f'"{filtered_id}",'
            else:
                filtered_ids_str += f"{filtered_id},"
        filtered_ids_str = filtered_ids_str[0:-1]

        if len(filtered_ids_str) > 0:
            if expr is not None and len(expr) > 0:
                filter_expr = f" and {self._pk_field_name} not in [{filtered_ids_str}]"
                return "(" + expr + ")" + filter_expr
            return f"{self._pk_field_name} not in [{filtered_ids_str}]"
        return expr

    def __next_params(self, coefficient: int):
        coefficient = max(1, coefficient)
        next_params = deepcopy(self._param)
        if metrics_positive_related(self._param[METRIC_TYPE]):
            next_radius = self._tail_band + self._width * coefficient
            if RADIUS in self._param[PARAMS] and next_radius > self._param[PARAMS][RADIUS]:
                next_params[PARAMS][RADIUS] = self._param[PARAMS][RADIUS]
            else:
                next_params[PARAMS][RADIUS] = next_radius
        else:
            next_radius = self._tail_band - self._width * coefficient
            if RADIUS in self._param[PARAMS] and next_radius < self._param[PARAMS][RADIUS]:
                next_params[PARAMS][RADIUS] = self._param[PARAMS][RADIUS]
            else:
                next_params[PARAMS][RADIUS] = next_radius
        next_params[PARAMS][RANGE_FILTER] = self._tail_band
        log.debug(
            f"next round search iteration radius:{next_params[PARAMS][RADIUS]},"
            f"range_filter:{next_params[PARAMS][RANGE_FILTER]},"
            f"coefficient:{coefficient}"
        )

        return next_params

    def close(self):
        iterator_cache.release_cache(self._cache_id)


class IteratorCache:
    def __init__(self) -> None:
        self._cache_id = 0
        self._cache_map = {}

    def cache(self, result: Any, cache_id: int):
        if cache_id == NO_CACHE_ID:
            self._cache_id += 1
            cache_id = self._cache_id
        self._cache_map[cache_id] = result
        return cache_id

    def fetch_cache(self, cache_id: int):
        return self._cache_map.get(cache_id, None)

    def release_cache(self, cache_id: int):
        if self._cache_map.get(cache_id, None) is not None:
            self._cache_map.pop(cache_id)


NO_CACHE_ID = -1
# Singleton Mode in Python
iterator_cache = IteratorCache()



# hello_text_embedding.py demonstrates how to insert raw data only into Milvus and perform
# dense vector based ANN search using TextEmbedding.
# 1. connect to Milvus
# 2. create collection
# 3. insert data
# 4. create index
# 5. search
# 6. drop collection
import time

from pymilvus import (
    connections,
    utility,
    FieldSchema, CollectionSchema, Function, DataType, FunctionType,
    Collection,
)

fmt = "\n=== {:30} ===\n"
search_latency_fmt = "search latency = {:.4f}s"

#################################################################################
# 1. connect to Milvus
# Add a new connection alias `default` for Milvus server in `localhost:19530`
print(fmt.format("start connecting to Milvus"))
connections.connect("default", host="localhost", port="19530")

collection_name = "text_embedding"

has = utility.has_collection(collection_name)
print(f"Does collection {collection_name} exist in Milvus: {has}")

#################################################################################
# 2. create collection
# We're going to create a collection with 2 explicit fields and a function.
# +-+------------+------------+------------------+------------------------------+
# | | field name | field type | other attributes |       field description      |
# +-+------------+------------+------------------+------------------------------+
# |1|    "id"    |   INT64    |  is_primary=True |      "primary field"         |
# | |            |            |   auto_id=False  |                              |
# +-+------------+------------+------------------+------------------------------+
# |2| "document" | VarChar    |                  |     "raw text document"      |
# +-+------------+------------+------------------+------------------------------+
#
# Function 'text embedding' is used to convert raw text document to a dense vector representation
# and store it in the 'dense' field.
# +-+------------+-------------------+-----------+------------------------------+
# | | field name | field type        | other attr|       field description      |
# +-+------------+-------------------+-----------+------------------------------+
# |3|  "dense"   |FLOAT_VECTOR       |  dim=1536 |                              |
# +-+------------+-------------------+-----------+------------------------------+
#
fields = [
    FieldSchema(name="id", dtype=DataType.INT64, is_primary=True, auto_id=True),
    FieldSchema(name="dense", dtype=DataType.FLOAT_VECTOR, dim=1536),
    FieldSchema(name="document", dtype=DataType.VARCHAR, max_length=1000),
]

text_embedding_function = Function(
    name="openai",
    function_type=FunctionType.TEXTEMBEDDING,
    input_field_names=["document"],
    output_field_names="dense",
    params={
        "provider": "openai",
        "model_name": "text-embedding-3-small",
    }
)

schema = CollectionSchema(fields, "hello_text_embedding demo")
schema.add_function(text_embedding_function)

hello_text_embedding = Collection(collection_name, schema, consistency_level="Strong")

print(fmt.format("Start inserting documents"))
docs = [
    "Artificial intelligence was founded as an academic discipline in 1956.",
    "Alan Turing was the first person to conduct substantial research in AI.",
    "Born in Maida Vale, London, Turing was raised in southern England.",
]

insert_result = hello_text_embedding.insert([docs])
ids = insert_result.primary_keys

################################################################################
# 4. create index
# We are going to create an index for collection, here we simply
# uses AUTOINDEX so Milvus can use the default parameters.
print(fmt.format("Start Creating index AUTOINDEX"))
index = {
    "index_type": "AUTOINDEX",
    "metric_type": "IP",
}

hello_text_embedding.create_index("dense", index)

################################################################################
# 5. search, query, and scalar filtering search
# After data were inserted into Milvus and indexed, you can perform:
# - search texts relevance by TextEmbedding using dense vector ANN search

# Before conducting a search or a query, you need to load the data into memory.
print(fmt.format("Start loading"))
hello_text_embedding.load()

# -----------------------------------------------------------------------------
search_params = {
    "metric_type": "IP",
    "params": {"nprobe": 10},
}
queries = ["When was artificial intelligence founded", 
           "Where was Alan Turing born?"]

start_time = time.time()
result = hello_text_embedding.search(queries, "dense", search_params, limit=3, output_fields=["document"], consistency_level="Strong")
end_time = time.time()

for hits, text in zip(result, queries):
    print(f"result of text: {text}")
    for hit in hits:
        print(f"\thit: {hit}, document field: {hit.entity.get('document')}")
print(search_latency_fmt.format(end_time - start_time))

# Finally, drop the collection
utility.drop_collection(collection_name)



# Copyright 2017 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base class for interceptors that operate on all RPC types."""

import time
from typing import Any, Callable, List, NamedTuple, Optional, Tuple

import grpc


def current_time_ms() -> str:
    return str(time.time_ns() // 1_000_000)


def _api_level_md(**kwargs) -> Optional[List[Tuple]]:
    metadata = []
    metadata.append(("client-request-unixmsec", current_time_ms()))
    client_request_id = kwargs.get("client-request-id", kwargs.get("client_request_id"))
    if client_request_id:
        metadata.append(("client-request-id", client_request_id))
    return metadata


class _GenericClientInterceptor(
    grpc.UnaryUnaryClientInterceptor,
    grpc.UnaryStreamClientInterceptor,
    grpc.StreamUnaryClientInterceptor,
    grpc.StreamStreamClientInterceptor,
):
    def __init__(self, interceptor_function: Callable) -> None:
        super().__init__()
        self._fn = interceptor_function

    def intercept_unary_unary(self, continuation: Callable, client_call_details: Any, request: Any):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, iter((request,))
        )
        response = continuation(new_details, next(new_request_iterator))
        return postprocess(response) if postprocess else response

    def intercept_unary_stream(
        self,
        continuation: Callable,
        client_call_details: Any,
        request: Any,
    ):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, iter((request,))
        )
        response_it = continuation(new_details, next(new_request_iterator))
        return postprocess(response_it) if postprocess else response_it

    def intercept_stream_unary(
        self,
        continuation: Callable,
        client_call_details: Any,
        request_iterator: Any,
    ):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, request_iterator
        )
        response = continuation(new_details, new_request_iterator)
        return postprocess(response) if postprocess else response

    def intercept_stream_stream(
        self,
        continuation: Callable,
        client_call_details: Any,
        request_iterator: Any,
    ):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, request_iterator
        )
        response_it = continuation(new_details, new_request_iterator)
        return postprocess(response_it) if postprocess else response_it


class ClientCallDetailsTuple(NamedTuple):
    method: Any
    timeout: Any
    metadata: Any
    credentials: Any


class _ClientCallDetails(ClientCallDetailsTuple, grpc.ClientCallDetails):
    pass


def header_adder_interceptor(headers: List, values: List):
    def intercept_call(
        client_call_details: Any,
        request_iterator: Any,
    ):
        metadata = []
        if client_call_details.metadata is not None:
            metadata = list(client_call_details.metadata)
        for item in zip(headers, values):
            metadata.append(item)
        client_call_details = _ClientCallDetails(
            client_call_details.method,
            client_call_details.timeout,
            metadata,
            client_call_details.credentials,
        )
        return client_call_details, request_iterator, None

    return _GenericClientInterceptor(intercept_call)



import logging
import time
from enum import IntEnum
from typing import Any, ClassVar, Dict, List, Optional, TypeVar, Union

import numpy as np
import orjson

from pymilvus.exceptions import (
    AutoIDException,
    ExceptionsMessage,
    InvalidConsistencyLevel,
)
from pymilvus.grpc_gen import common_pb2, rg_pb2, schema_pb2
from pymilvus.grpc_gen import milvus_pb2 as milvus_types

from . import utils

Status = TypeVar("Status")
ConsistencyLevel = common_pb2.ConsistencyLevel

logger = logging.getLogger(__name__)

ALWAYS_KEEP_ZERO_KEYS = frozenset(
    {"scanned_remote_bytes", "scanned_total_bytes", "cache_hit_ratio"}
)


# OmitZeroDict: ignore the key-value pairs with value as 0 when printing
class OmitZeroDict(dict):
    def omit_zero_len(self):
        return len({k: v for k, v in self.items() if v or k in ALWAYS_KEEP_ZERO_KEYS})

    # keep zero for specific keys, omit other zero values
    def __str__(self):
        return str({k: v for k, v in self.items() if v or k in ALWAYS_KEEP_ZERO_KEYS})

    # no filter
    def __repr__(self):
        return str(dict(self))


class Status:
    """
    :attribute code: int (optional) default as ok

    :attribute message: str (optional) current status message
    """

    SUCCESS = 0
    UNEXPECTED_ERROR = 1
    CONNECT_FAILED = 2
    PERMISSION_DENIED = 3
    COLLECTION_NOT_EXISTS = 4
    ILLEGAL_ARGUMENT = 5
    ILLEGAL_RANGE = 6
    ILLEGAL_DIMENSION = 7
    # TODO in new error code, 8 is for RATE_LIMIT, 9 for FORCE_DENY
    ILLEGAL_INDEX_TYPE = 8
    ILLEGAL_COLLECTION_NAME = 9
    ILLEGAL_TOPK = 10
    ILLEGAL_ROWRECORD = 11
    ILLEGAL_VECTOR_ID = 12
    ILLEGAL_SEARCH_RESULT = 13
    FILE_NOT_FOUND = 14
    META_FAILED = 15
    CACHE_FAILED = 16
    CANNOT_CREATE_FOLDER = 17
    CANNOT_CREATE_FILE = 18
    CANNOT_DELETE_FOLDER = 19
    CANNOT_DELETE_FILE = 20
    BUILD_INDEX_ERROR = 21
    ILLEGAL_NLIST = 22
    ILLEGAL_METRIC_TYPE = 23
    OUT_OF_MEMORY = 24
    INDEX_NOT_EXIST = 25
    EMPTY_COLLECTION = 26

    def __init__(self, code: int = SUCCESS, message: str = "Success") -> None:
        self.code = code
        self.message = message

    def __repr__(self) -> str:
        attr_list = [f"{key}={value}" for key, value in self.__dict__.items()]
        return f"{self.__class__.__name__}({', '.join(attr_list)})"

    def __eq__(self, other: Union[int, Status]):
        """Make Status comparable with self by code"""
        if isinstance(other, int):
            return self.code == other

        return isinstance(other, self.__class__) and self.code == other.code

    def OK(self):
        return self.code == Status.SUCCESS


class DataType(IntEnum):
    """
    String of DataType is str of its value, e.g.: str(DataType.BOOL) == "1"
    """

    NONE = 0  # schema_pb2.None, this is an invalid representation in python
    BOOL = schema_pb2.Bool
    INT8 = schema_pb2.Int8
    INT16 = schema_pb2.Int16
    INT32 = schema_pb2.Int32
    INT64 = schema_pb2.Int64

    FLOAT = schema_pb2.Float
    DOUBLE = schema_pb2.Double

    STRING = schema_pb2.String
    VARCHAR = schema_pb2.VarChar
    ARRAY = schema_pb2.Array
    JSON = schema_pb2.JSON
    GEOMETRY = schema_pb2.Geometry

    BINARY_VECTOR = schema_pb2.BinaryVector
    FLOAT_VECTOR = schema_pb2.FloatVector
    FLOAT16_VECTOR = schema_pb2.Float16Vector
    BFLOAT16_VECTOR = schema_pb2.BFloat16Vector
    SPARSE_FLOAT_VECTOR = schema_pb2.SparseFloatVector
    INT8_VECTOR = schema_pb2.Int8Vector

    STRUCT = schema_pb2.Struct

    # Internal use only - not exposed to users
    _ARRAY_OF_VECTOR = schema_pb2.ArrayOfVector
    _ARRAY_OF_STRUCT = schema_pb2.ArrayOfStruct

    UNKNOWN = 999

    def __str__(self) -> str:
        return str(self.value)


class FunctionType(IntEnum):
    UNKNOWN = 0
    BM25 = 1
    TEXTEMBEDDING = 2
    RERANK = 3


class RangeType(IntEnum):
    LT = 0  # less than
    LTE = 1  # less than or equal
    EQ = 2  # equal
    GT = 3  # greater than
    GTE = 4  # greater than or equal
    NE = 5  # not equal


class IndexType(IntEnum):
    INVALID = 0
    FLAT = 1
    IVFLAT = 2
    IVF_SQ8 = 3
    RNSG = 4
    IVF_SQ8H = 5
    IVF_PQ = 6
    HNSW = 11
    ANNOY = 12

    # alternative name
    IVF_FLAT = IVFLAT
    IVF_SQ8_H = IVF_SQ8H

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self._name_}>"

    def __str__(self) -> str:
        return self._name_


class MetricType(IntEnum):
    INVALID = 0
    L2 = 1
    IP = 2
    # Only supported for byte vectors
    HAMMING = 3
    JACCARD = 4
    TANIMOTO = 5
    SUBSTRUCTURE = 6
    SUPERSTRUCTURE = 7

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self._name_}>"

    def __str__(self) -> str:
        return self._name_


class IndexState(IntEnum):
    IndexStateNone = 0
    Unissued = 1
    InProgress = 2
    Finished = 3
    Failed = 4
    Deleted = 5


class PlaceholderType(IntEnum):
    NoneType = 0
    BinaryVector = 100
    FloatVector = 101
    FLOAT16_VECTOR = 102
    BFLOAT16_VECTOR = 103
    SparseFloatVector = 104
    Int8Vector = 105
    VARCHAR = 21

    EmbListBinaryVector = 300
    EmbListFloatVector = 301
    EmbListFloat16Vector = 302
    EmbListBFloat16Vector = 303
    EmbListSparseFloatVector = 304
    EmbListInt8Vector = 305


class State(IntEnum):
    """
    UndefiedState:  Unknown
    Executing:      indicating this compaction has undone plans.
    Completed:      indicating all the plans of this compaction are done,
                    no matter successful or not.
    """

    UndefiedState = 0
    Executing = 1
    Completed = 2

    @staticmethod
    def new(s: int):
        if s == State.Executing:
            return State.Executing
        if s == State.Completed:
            return State.Completed
        return State.UndefiedState

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self._name_}>"

    def __str__(self) -> str:
        return self._name_


class LoadState(IntEnum):
    """
    NotExist: collection or partition isn't existed
    NotLoad:  collection or partition isn't loaded
    Loading:  collection or partition is loading
    Loaded:   collection or partition is loaded
    """

    NotExist = 0
    NotLoad = 1
    Loading = 2
    Loaded = 3

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self._name_}>"

    def __str__(self) -> str:
        return self._name_


class CompactionState:
    """
    in_executing: number of plans in executing
    in_timeout:   number of plans failed of timeout
    completed:    number of plans successfully completed
    """

    def __init__(
        self,
        compaction_id: int,
        state: State,
        in_executing: int,
        in_timeout: int,
        completed: int,
    ) -> None:
        self.compaction_id = compaction_id
        self.state = state
        self.in_executing = in_executing
        self.in_timeout = in_timeout
        self.completed = completed

    @property
    def state_name(self):
        return self.state.name

    def __repr__(self) -> str:
        return f"""
CompactionState
 - compaction id: {self.compaction_id}
 - State: {self.state}
 - executing plan number: {self.in_executing}
 - timeout plan number: {self.in_timeout}
 - complete plan number: {self.completed}
"""


class Plan:
    def __init__(self, sources: list, target: int) -> None:
        self.sources = sources
        self.target = target

    def __repr__(self) -> str:
        return f"""
Plan:
 - sources: {self.sources}
 - target: {self.target}
"""


class CompactionPlans:
    def __init__(self, compaction_id: int, state: int) -> None:
        self.compaction_id = compaction_id
        self.state = State.new(state)
        self.plans = []

    def __repr__(self) -> str:
        return f"""
Compaction Plans:
 - compaction id: {self.compaction_id}
 - state: {self.state}
 - plans: {self.plans}
 """


def cmp_consistency_level(l1: Union[str, int], l2: Union[str, int]):
    if isinstance(l1, str):
        try:
            l1 = ConsistencyLevel.Value(l1)
        except ValueError:
            return False

    if isinstance(l2, str):
        try:
            l2 = ConsistencyLevel.Value(l2)
        except ValueError:
            return False

    if isinstance(l1, int) and l1 not in ConsistencyLevel.values():
        return False

    if isinstance(l2, int) and l2 not in ConsistencyLevel.values():
        return False

    return l1 == l2


def get_consistency_level(consistency_level: Union[str, int]):
    if isinstance(consistency_level, int):
        if consistency_level in ConsistencyLevel.values():
            return consistency_level
        raise InvalidConsistencyLevel(message=f"invalid consistency level: {consistency_level}")
    if isinstance(consistency_level, str):
        try:
            return ConsistencyLevel.Value(consistency_level)
        except ValueError as e:
            raise InvalidConsistencyLevel(
                message=f"invalid consistency level: {consistency_level}"
            ) from e
    raise InvalidConsistencyLevel(message="invalid consistency level")


class Shard:
    def __init__(self, channel_name: str, shard_nodes: list, shard_leader: int) -> None:
        self._channel_name = channel_name
        self._shard_nodes = set(shard_nodes)
        self._shard_leader = shard_leader

    def __repr__(self) -> str:
        return (
            f"Shard: <channel_name:{self.channel_name}>, "
            f"<shard_leader:{self.shard_leader}>, <shard_nodes:{self.shard_nodes}>"
        )

    @property
    def channel_name(self) -> str:
        return self._channel_name

    @property
    def shard_nodes(self):
        return self._shard_nodes

    @property
    def shard_leader(self) -> int:
        return self._shard_leader


class Group:
    """
    This class represents replica info in orm format api, which is deprecated in milvus client api.
    use `ReplicaInfo` instead.
    """

    def __init__(
        self,
        group_id: int,
        shards: List[str],
        group_nodes: List[tuple],
        resource_group: str,
        num_outbound_node: dict,
    ) -> None:
        self._id = group_id
        self._shards = shards
        self._group_nodes = tuple(group_nodes)
        self._resource_group = resource_group
        self._num_outbound_node = num_outbound_node

    def __repr__(self) -> str:
        return (
            f"Group: <group_id:{self.id}>, <group_nodes:{self.group_nodes}>, "
            f"<shards:{self.shards}>, <resource_group: {self.resource_group}>, "
            f"<num_outbound_node: {self.num_outbound_node}>"
        )

    @property
    def id(self):
        return self._id

    @property
    def group_nodes(self):
        return self._group_nodes

    @property
    def shards(self):
        return self._shards

    @property
    def resource_group(self):
        return self._resource_group

    @property
    def num_outbound_node(self):
        return self._num_outbound_node


class Replica:
    """
    This class represents replica info list in orm format api,
    which is deprecated in milvus client api.
    use `List[ReplicaInfo]` instead.

    Replica groups:
        - Group: <group_id:2>, <group_nodes:(1, 2, 3)>,
            <shards:[Shard: <shard_id:10>,
                <channel_name:channel-1>,
                <shard_leader:1>,
                <shard_nodes:(1, 2, 3)>]>
        - Group: <group_id:2>, <group_nodes:(1, 2, 3)>,
            <shards:[Shard:
                <shard_id:10>,
                <channel_name:channel-1>,
                <shard_leader:1>,
                <shard_nodes:(1, 2, 3)>]>
    """

    def __init__(self, groups: list) -> None:
        self._groups = groups

    def __repr__(self) -> str:
        s = "Replica groups:"
        for g in self.groups:
            s += f"\n- {g}"
        return s

    @property
    def groups(self):
        return self._groups


class ReplicaInfo:
    def __init__(
        self,
        replica_id: int,
        shards: List[str],
        nodes: List[tuple],
        resource_group: str,
        num_outbound_node: dict,
    ) -> None:
        self._id = replica_id
        self._shards = shards
        self._nodes = tuple(nodes)
        self._resource_group = resource_group
        self._num_outbound_node = num_outbound_node

    def __repr__(self) -> str:
        return (
            f"ReplicaInfo: <id:{self.id}>, <nodes:{self.group_nodes}>, "
            f"<shards:{self.shards}>, <resource_group: {self.resource_group}>, "
            f"<num_outbound_node: {self.num_outbound_node}>"
        )

    @property
    def id(self):
        return self._id

    @property
    def group_nodes(self):
        return self._nodes

    @property
    def shards(self):
        return self._shards

    @property
    def resource_group(self):
        return self._resource_group

    @property
    def num_outbound_node(self):
        return self._num_outbound_node


class BulkInsertState:
    """enum states of bulk insert task"""

    ImportPending = 0
    ImportFailed = 1
    ImportStarted = 2
    ImportPersisted = 5
    ImportCompleted = 6
    ImportFailedAndCleaned = 7
    ImportUnknownState = 100

    """pre-defined keys of bulk insert task info"""
    FAILED_REASON = "failed_reason"
    IMPORT_FILES = "files"
    IMPORT_COLLECTION = "collection"
    IMPORT_PARTITION = "partition"
    IMPORT_PROGRESS = "progress_percent"

    """
    Bulk insert state example:
        - taskID    : 44353845454358,
        - state     : "BulkLoadPersisted",
        - row_count : 1000,
        - infos     : {"files": "rows.json",
                       "collection": "c1",
                       "partition": "",
                       "failed_reason": ""},
        - id_list   : [44353845455401, 44353845456401]
        - create_ts : 1661398759,
    """

    state_2_state: ClassVar[Dict] = {
        common_pb2.ImportPending: ImportPending,
        common_pb2.ImportFailed: ImportFailed,
        common_pb2.ImportStarted: ImportStarted,
        common_pb2.ImportPersisted: ImportPersisted,
        common_pb2.ImportCompleted: ImportCompleted,
        common_pb2.ImportFailedAndCleaned: ImportFailedAndCleaned,
    }

    state_2_name: ClassVar[Dict] = {
        ImportPending: "Pending",
        ImportFailed: "Failed",
        ImportStarted: "Started",
        ImportPersisted: "Persisted",
        ImportCompleted: "Completed",
        ImportFailedAndCleaned: "Failed and cleaned",
        ImportUnknownState: "Unknown",
    }

    def __init__(
        self,
        task_id: int,
        state: State,
        row_count: int,
        id_ranges: list,
        infos: Dict,
        create_ts: int,
    ):
        self._task_id = task_id
        self._state = state
        self._row_count = row_count
        self._id_ranges = id_ranges
        self._create_ts = create_ts

        self._infos = {kv.key: kv.value for kv in infos}

    def __repr__(self) -> str:
        fmt = """<Bulk insert state:
    - taskID          : {},
    - state           : {},
    - row_count       : {},
    - infos           : {},
    - id_ranges       : {},
    - create_ts       : {}
>"""
        return fmt.format(
            self._task_id,
            self.state_name,
            self.row_count,
            self.infos,
            self.id_ranges,
            self.create_time_str,
        )

    @property
    def task_id(self):
        """
        Return unique id of this task.
        """
        return self._task_id

    @property
    def row_count(self):
        """
        If the task is finished, this value is the number of rows imported.
        If the task is not finished, this value is the number of rows parsed.
        """
        return self._row_count

    @property
    def state(self):
        return self.state_2_state.get(self._state, BulkInsertState.ImportUnknownState)

    @property
    def state_name(self) -> str:
        return self.state_2_name.get(self._state, "unknown state")

    @property
    def id_ranges(self):
        """
        auto generated id ranges if the primary key is auto generated

        the id list of response is id ranges
        for example, if the response return [1, 100, 200, 250]
        the full id list should be [1, 2, 3 ... , 99, 100, 200, 201, 202 ... , 249, 250]
        """
        return self._id_ranges

    @property
    def ids(self):
        """
        auto generated ids if the primary key is auto generated

        the id list of response is id ranges
        for example, if the response return [1, 100, 200, 250], the id ranges: [1,100),[200,250)
        the full id list should be [1, 2, 3 ... , 99, 200, 201, 202 ... , 249]
        """

        if len(self._id_ranges) % 2 != 0:
            raise AutoIDException(message=ExceptionsMessage.AutoIDIllegalRanges)

        ids = []
        for i in range(int(len(self._id_ranges) / 2)):
            begin = self._id_ranges[i * 2]
            end = self._id_ranges[i * 2 + 1]
            for j in range(begin, end):
                ids.append(j)

        return ids

    @property
    def infos(self):
        """more informations about the task, progress percentage, file path, failed reason, etc."""
        return self._infos

    @property
    def failed_reason(self):
        """failed reason of the bulk insert task."""
        return self._infos.get(BulkInsertState.FAILED_REASON, "")

    @property
    def files(self):
        """data files of the bulk insert task."""
        return self._infos.get(BulkInsertState.IMPORT_FILES, "")

    @property
    def collection_name(self):
        """target collection's name of the bulk insert task."""
        return self._infos.get(BulkInsertState.IMPORT_COLLECTION, "")

    @property
    def partition_name(self):
        """target partition's name of the bulk insert task."""
        return self._infos.get(BulkInsertState.IMPORT_PARTITION, "")

    @property
    def create_timestamp(self):
        """the integer timestamp when this task is created."""
        return self._create_ts

    @property
    def create_time_str(self):
        """A readable string converted from the timestamp when this task is created."""
        ts = time.localtime(self._create_ts)
        return time.strftime("%Y-%m-%d %H:%M:%S", ts)

    @property
    def progress(self):
        """working progress percent value."""
        percent = self._infos.get(BulkInsertState.IMPORT_PROGRESS, "0")
        return int(percent)


class GrantItem:
    def __init__(self, entity: Any) -> None:
        self._object = entity.object.name
        self._object_name = entity.object_name
        self._db_name = entity.db_name
        self._role_name = entity.role.name
        self._grantor_name = entity.grantor.user.name
        self._privilege = entity.grantor.privilege.name

    def __repr__(self) -> str:
        return (
            f"GrantItem: <object:{self.object}>, <object_name:{self.object_name}>, "
            f"<db_name:{self.db_name}>, "
            f"<role_name:{self.role_name}>, <grantor_name:{self.grantor_name}>, "
            f"<privilege:{self.privilege}>"
        )

    @property
    def object(self):
        return self._object

    @property
    def object_name(self):
        return self._object_name

    @property
    def db_name(self):
        return self._db_name

    @property
    def role_name(self):
        return self._role_name

    @property
    def grantor_name(self):
        return self._grantor_name

    @property
    def privilege(self):
        return self._privilege

    def __iter__(self):
        yield "object_type", self.object
        yield "object_name", self.object_name
        if self.db_name:
            yield "db_name", self.db_name

        yield "role_name", self.role_name
        yield "privilege", self.privilege
        if self.grantor_name:
            yield "grantor_name", self.grantor_name


class GrantInfo:
    """
    GrantInfo groups:
    - GrantItem: <object:Collection>, <object_name:foo>, <role_name:x>,
        <grantor_name:root>, <privilege:Load>
    - GrantItem: <object:Global>, <object_name:*>, <role_name:x>,
        <grantor_name:root>, <privilege:CreateCollection>
    """

    def __init__(self, entities: List[milvus_types.RoleEntity]) -> None:
        groups = []
        for entity in entities:
            if isinstance(entity, milvus_types.GrantEntity):
                groups.append(GrantItem(entity))

        self._groups = groups

    def __repr__(self) -> str:
        s = "GrantInfo groups:"
        for g in self.groups:
            s += f"\n- {g}"
        return s

    @property
    def groups(self):
        return self._groups


class PrivilegeGroupItem:
    def __init__(self, privilege_group: str, privileges: List[milvus_types.PrivilegeEntity]):
        self._privilege_group = privilege_group
        privielges = []
        for privilege in privileges:
            if isinstance(privilege, milvus_types.PrivilegeEntity):
                privielges.append(privilege.name)
        self._privileges = tuple(privielges)

    def __repr__(self) -> str:
        return f"PrivilegeGroupItem: <privilege_group:{self.privilege_group}>, <privileges:{self.privileges}>"

    @property
    def privilege_group(self):
        return self._privilege_group

    @property
    def privileges(self):
        return self._privileges


class PrivilegeGroupInfo:
    """
    PrivilegeGroupInfo groups:
    - PrivilegeGroupItem: <privilege_group:group>, <privileges:('Load', 'CreateCollection')>
    """

    def __init__(self, results: List[milvus_types.PrivilegeGroupInfo]) -> None:
        groups = []
        for result in results:
            if isinstance(result, milvus_types.PrivilegeGroupInfo):
                groups.append(PrivilegeGroupItem(result.group_name, result.privileges))
        self._groups = groups

    def __repr__(self) -> str:
        s = "PrivilegeGroupInfo groups:"
        for g in self.groups:
            s += f"\n- {g}"
        return s

    @property
    def groups(self):
        return self._groups


class UserItem:
    def __init__(self, username: str, entities: List[milvus_types.RoleEntity]) -> None:
        self._username = username
        roles = []
        for entity in entities:
            if isinstance(entity, milvus_types.RoleEntity):
                roles.append(entity.name)
        self._roles = tuple(roles)

    def __repr__(self) -> str:
        return f"UserItem: <username:{self.username}>, <roles:{self.roles}>"

    @property
    def username(self):
        return self._username

    @property
    def roles(self):
        return self._roles


class UserInfo:
    """
    UserInfo groups:
    - UserItem: <username:root>, <roles:('admin', 'public')>
    """

    def __init__(self, results: List[milvus_types.UserResult]):
        groups = []
        for result in results:
            if isinstance(result, milvus_types.UserResult):
                groups.append(UserItem(result.user.name, result.roles))

        self._groups = groups

    def __repr__(self) -> str:
        s = "UserInfo groups:"
        for g in self.groups:
            s += f"\n- {g}"
        return s

    @property
    def groups(self):
        return self._groups


class RoleItem:
    def __init__(self, role_name: str, entities: List[milvus_types.UserEntity]):
        self._role_name = role_name
        users = []
        for entity in entities:
            if isinstance(entity, milvus_types.UserEntity):
                users.append(entity.name)
        self._users = tuple(users)

    def __repr__(self) -> str:
        return f"RoleItem: <role_name:{self.role_name}>, <users:{self.users}>"

    @property
    def role_name(self):
        return self._role_name

    @property
    def users(self):
        return self._users


class RoleInfo:
    """
    RoleInfo groups:
    - UserItem: <role_name:admin>, <users:('root',)>
    """

    def __init__(self, results: List[milvus_types.RoleResult]) -> None:
        groups = []
        for result in results:
            if isinstance(result, milvus_types.RoleResult):
                groups.append(RoleItem(result.role.name, result.users))

        self._groups = groups

    def __repr__(self) -> str:
        s = "RoleInfo groups:"
        for g in self.groups:
            s += f"\n- {g}"
        return s

    @property
    def groups(self):
        return self._groups


class ResourceGroupInfo:
    def __init__(self, resource_group: Any) -> None:
        self._name = resource_group.name
        self._capacity = resource_group.capacity
        self._num_available_node = resource_group.num_available_node
        self._num_loaded_replica = resource_group.num_loaded_replica
        self._num_outgoing_node = resource_group.num_outgoing_node
        self._num_incoming_node = resource_group.num_incoming_node
        self._config = resource_group.config
        self._nodes = [NodeInfo(node) for node in resource_group.nodes]

    def __repr__(self) -> str:
        return f"""ResourceGroupInfo:
<name:{self.name}>,
<capacity:{self.capacity}>,
<num_available_node:{self.num_available_node}>,
<num_loaded_replica:{self.num_loaded_replica}>,
<num_outgoing_node:{self.num_outgoing_node}>,
<num_incoming_node:{self.num_incoming_node}>,
<config:{self.config}>,
<nodes:{self.nodes}>"""

    @property
    def name(self):
        return self._name

    @property
    def capacity(self):
        return self._capacity

    @property
    def num_available_node(self):
        return self._num_available_node

    @property
    def num_loaded_replica(self):
        return self._num_loaded_replica

    @property
    def num_outgoing_node(self):
        return self._num_outgoing_node

    @property
    def num_incoming_node(self):
        return self._num_incoming_node

    @property
    def config(self):
        return self._config

    @property
    def nodes(self):
        return self._nodes


class NodeInfo:
    """
    Represents information about a node in the system.
    Attributes:
        node_id (int): The ID of the node.
        address (str): The ip address of the node.
        hostname (str): The hostname of the node.
    Example:
        NodeInfo(
            node_id=1,
            address="127.0.0.1",
            hostname="localhost",
        )
    """

    def __init__(self, info: Any) -> None:
        self._node_id = info.node_id
        self._address = info.address
        self._hostname = info.hostname

    def __repr__(self) -> str:
        return f"""NodeInfo:
<node_id:{self.node_id}>,
<address:{self.address}>,
<hostname:{self.hostname}>"""

    @property
    def node_id(self) -> int:
        return self._node_id

    @property
    def address(self) -> str:
        return self._address

    @property
    def hostname(self) -> str:
        return self._hostname


ResourceGroupConfig = rg_pb2.ResourceGroupConfig
"""
Represents the configuration of a resource group.
Attributes:
    requests (ResourceGroupLimit): The requests of the resource group.
    limits (ResourceGroupLimit): The limits of the resource group.
    transfer_from (List[ResourceGroupTransfer]): The transfer config that resource group
        can transfer node from the resource group of this field at high priority.
    transfer_to (List[ResourceGroupTransfer]): The transfer config that resource group
        can transfer node to the resource group of this field at high priority.
Example:
    ResourceGroupConfig(
        requests={"node_num": 1},
        limits={"node_num": 5},
        transfer_from=[{"resource_group": "__default_resource_group"}],
        transfer_to=[{"resource_group": "resource_group_2"}],
    )
"""

ResourceGroupLimit = rg_pb2.ResourceGroupLimit
"""
Represents the limit of a resource group.
Attributes:
    node_num (int): The number of nodes that the resource group can hold.
"""

ResourceGroupTransfer = rg_pb2.ResourceGroupTransfer
"""
Represents the transfer config of a resource group.
Attributes:
    resource_group (str): The name of the resource group that can be transferred to or from.
"""


class HybridExtraList(list):
    """
    A list that holds partially eager and partially lazy row data.

    - Primitive fields are extracted at initialization.
    - Variable-length fields (array/string/json) are extracted lazily on access.

    Attributes:
        extra (dict): Extra metadata associated with the result set.
    """

    def __init__(
        self,
        lazy_field_data: List[Any],  # lazy extract fields
        *args,
        extra: Optional[Dict] = None,
        dynamic_fields: Optional[List] = None,
        strict_float32: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._lazy_field_data = lazy_field_data
        self._dynamic_fields = dynamic_fields
        self.extra = OmitZeroDict(extra or {})
        self._float_vector_np_array = {}
        self._has_materialized_float_vector = False
        self._strict_float32 = strict_float32
        self._materialized_bitmap = [False] * len(self)

    def _extract_lazy_fields(self, index: int, field_data: Any, row_data: Dict) -> Any:
        if field_data.type == DataType.JSON:
            if len(field_data.valid_data) > 0 and field_data.valid_data[index] is False:
                row_data[field_data.field_name] = None
                return
            try:
                json_dict = orjson.loads(field_data.scalars.json_data.data[index])
            except Exception as e:
                logger.error(
                    f"HybridExtraList::_extract_lazy_fields::Failed to load JSON data: {e}, original data: {field_data.scalars.json_data.data[index]}"
                )
                raise
            if not field_data.is_dynamic:
                row_data[field_data.field_name] = json_dict
                return
            if not self._dynamic_fields:
                # Only update keys that don't exist in row_data
                row_data.update({k: v for k, v in json_dict.items() if k not in row_data})
                return
            # Only update keys that don't exist in row_data and are in dynamic_fields
            row_data.update(
                {
                    k: v
                    for k, v in json_dict.items()
                    if k in self._dynamic_fields and k not in row_data
                }
            )
        elif field_data.type == DataType.FLOAT_VECTOR:
            dim = field_data.vectors.dim
            start_pos = index * dim
            end_pos = start_pos + dim
            if len(field_data.vectors.float_vector.data) >= start_pos:
                # Here we use numpy.array to convert the float64 values to numpy.float32 values,
                # and return a list of numpy.float32 to users
                # By using numpy.array, performance improved by 60% for topk=16384 dim=1536 case.
                if self._strict_float32:
                    row_data[field_data.field_name] = self._float_vector_np_array[
                        field_data.field_name
                    ][start_pos:end_pos]
                else:
                    row_data[field_data.field_name] = field_data.vectors.float_vector.data[
                        start_pos:end_pos
                    ]
        elif field_data.type == DataType.BINARY_VECTOR:
            dim = field_data.vectors.dim
            bytes_per_vector = dim // 8
            start_pos = index * bytes_per_vector
            end_pos = start_pos + bytes_per_vector
            if len(field_data.vectors.binary_vector) >= start_pos:
                row_data[field_data.field_name] = [
                    field_data.vectors.binary_vector[start_pos:end_pos]
                ]
        elif field_data.type == DataType.BFLOAT16_VECTOR:
            dim = field_data.vectors.dim
            bytes_per_vector = dim * 2
            start_pos = index * bytes_per_vector
            end_pos = start_pos + bytes_per_vector
            if len(field_data.vectors.bfloat16_vector) >= start_pos:
                row_data[field_data.field_name] = [
                    field_data.vectors.bfloat16_vector[start_pos:end_pos]
                ]
        elif field_data.type == DataType.FLOAT16_VECTOR:
            dim = field_data.vectors.dim
            bytes_per_vector = dim * 2
            start_pos = index * bytes_per_vector
            end_pos = start_pos + bytes_per_vector
            if len(field_data.vectors.float16_vector) >= start_pos:
                row_data[field_data.field_name] = [
                    field_data.vectors.float16_vector[start_pos:end_pos]
                ]
        elif field_data.type == DataType.SPARSE_FLOAT_VECTOR:
            row_data[field_data.field_name] = utils.sparse_parse_single_row(
                field_data.vectors.sparse_float_vector.contents[index]
            )
        elif field_data.type == DataType.INT8_VECTOR:
            dim = field_data.vectors.dim
            start_pos = index * dim
            end_pos = start_pos + dim
            if len(field_data.vectors.int8_vector) >= start_pos:
                row_data[field_data.field_name] = [
                    field_data.vectors.int8_vector[start_pos:end_pos]
                ]
        elif field_data.type == DataType._ARRAY_OF_VECTOR:
            # Handle array of vectors
            if hasattr(field_data, "vectors") and hasattr(field_data.vectors, "vector_array"):
                if index < len(field_data.vectors.vector_array.data):
                    vector_data = field_data.vectors.vector_array.data[index]
                    dim = vector_data.dim
                    float_data = vector_data.float_vector.data
                    num_vectors = len(float_data) // dim
                    row_vectors = []
                    for vec_idx in range(num_vectors):
                        vec_start = vec_idx * dim
                        vec_end = vec_start + dim
                        row_vectors.append(list(float_data[vec_start:vec_end]))
                    row_data[field_data.field_name] = row_vectors
                else:
                    row_data[field_data.field_name] = []
            else:
                row_data[field_data.field_name] = []
        elif field_data.type == DataType._ARRAY_OF_STRUCT:
            # Handle struct arrays - convert column format back to array of structs
            if hasattr(field_data, "struct_arrays") and field_data.struct_arrays:
                # Import here to avoid circular imports
                from .entity_helper import extract_struct_array_from_column_data  # noqa: PLC0415

                row_data[field_data.field_name] = extract_struct_array_from_column_data(
                    field_data.struct_arrays, index
                )
            else:
                row_data[field_data.field_name] = None

    def __getitem__(self, index: Union[int, slice]):
        if isinstance(index, slice):
            results = []
            for i in range(*index.indices(len(self))):
                row = self[i]
                results.append(row)
            return results

        if self._materialized_bitmap[index]:
            return super().__getitem__(index)

        self._pre_materialize_float_vector()
        row = super().__getitem__(index)
        for field_data in self._lazy_field_data:
            self._extract_lazy_fields(index, field_data, row)

        self._materialized_bitmap[index] = True
        return row

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def __str__(self) -> str:
        preview = [str(self[i]) for i in range(min(10, len(self)))]
        return f"data: {preview}{' ...' if len(self) > 10 else ''}, extra_info: {self.extra}"

    def _pre_materialize_float_vector(self):
        if not self._strict_float32 or self._has_materialized_float_vector:
            return
        for field_data in self._lazy_field_data:
            if field_data.type == DataType.FLOAT_VECTOR:
                self._float_vector_np_array[field_data.field_name] = np.array(
                    field_data.vectors.float_vector.data, dtype=np.float32
                )
        self._has_materialized_float_vector = True

    def materialize(self):
        """Materializes all lazy-loaded fields for all rows."""
        for i in range(len(self)):
            # By simply accessing the item, __getitem__ will trigger
            # the one-time materialization logic if it hasn't been done yet.
            _ = self[i]
        return self

    __repr__ = __str__


class ExtraList(list):
    """
    A list that can hold extra information.
    Attributes:
        extra (dict): The extra information of the list.
    Example:
        ExtraList([1, 2, 3], extra={"total": 3})
    """

    def __init__(self, *args, extra: Optional[Dict] = None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.extra = OmitZeroDict(extra or {})

    def __str__(self) -> str:
        """Only print at most 10 query results"""
        if self.extra and self.extra.omit_zero_len() != 0:
            return f"data: {list(map(str, self[:10]))}{' ...' if len(self) > 10 else ''}, extra_info: {self.extra}"
        return f"data: {list(map(str, self[:10]))}{' ...' if len(self) > 10 else ''}"

    __repr__ = __str__


def get_cost_from_status(status: Optional[common_pb2.Status] = None):
    return int(
        status.extra_info["report_value"]
        if status and status.extra_info and "report_value" in status.extra_info
        else "0"
    )


def get_extra_info(status: Optional[common_pb2.Status] = None):
    extra = {"cost": get_cost_from_status(status)}
    if status and status.extra_info:
        if "scanned_remote_bytes" in status.extra_info:
            extra["scanned_remote_bytes"] = int(status.extra_info["scanned_remote_bytes"])
        if "scanned_total_bytes" in status.extra_info:
            extra["scanned_total_bytes"] = int(status.extra_info["scanned_total_bytes"])
        if "cache_hit_ratio" in status.extra_info:
            extra["cache_hit_ratio"] = float(status.extra_info["cache_hit_ratio"])
    return extra


class DatabaseInfo:
    """
    Represents the information of a database.
    Atributes:
        name (str): The name of the database.
        properties (dict): The properties of the database.
    Example:
        DatabaseInfo(name="test_db", id=1, properties={"key": "value"})
    """

    @property
    def name(self) -> str:
        return self._name

    @property
    def properties(self) -> Dict:
        return self._properties

    def __init__(self, info: Any) -> None:
        self._name = info.db_name
        self._properties = {}

        for p in info.properties:
            self.properties[p.key] = p.value

    def __str__(self) -> str:
        return f"DatabaseInfo(name={self.name}, properties={self.properties})"

    def to_dict(self) -> Dict[str, Any]:
        """Converts the DatabaseInfo instance to a dictionary."""
        result = {"name": self.name}
        result.update(self.properties)
        return result


class AnalyzeToken:
    def __init__(
        self, token: milvus_types.AnalyzerToken, with_hash: bool = False, with_detail: bool = False
    ):
        self.dict = {"token": token.token}
        if with_detail:
            self.dict["start_offset"] = token.start_offset
            self.dict["end_offset"] = token.end_offset
            self.dict["position"] = token.position
            self.dict["position_length"] = token.position_length
        if with_hash:
            self.dict["hash"] = token.hash

    @property
    def token(self):
        return self.dict["token"]

    @property
    def start_offset(self):
        return self.dict["start_offset"]

    @property
    def end_offset(self):
        return self.dict["end_offset"]

    @property
    def position(self):
        return self.dict["position"]

    @property
    def position_length(self):
        return self.dict["position_length"]

    @property
    def hash(self):
        return self.dict["hash"]

    def __getitem__(self, key: str):
        return self.dict[key]

    def __str__(self):
        return str(self.dict)

    __repr__ = __str__


class AnalyzeResult:
    def __init__(
        self, info: milvus_types.AnalyzerResult, with_hash: bool = False, with_detail: bool = False
    ) -> None:
        if not with_detail and not with_hash:
            self.tokens = [token.token for token in info.tokens]
        else:
            self.tokens = [AnalyzeToken(token, with_hash, with_detail) for token in info.tokens]

    def __str__(self) -> str:
        return str(self.tokens)

    __repr__ = __str__



import numpy as np
from pymilvus import (
    MilvusClient,
    DataType
)

milvus_client = MilvusClient("http://localhost:19530")

db1Name = "db1"
# create db1
if db1Name not in milvus_client.list_databases():
    print("\ncreate database: db1")
    milvus_client.create_database(db_name=db1Name, properties={"key1":"value1"})
    db_info = milvus_client.describe_database(db_name=db1Name)
    print(db_info)


# alter_database_properties of db1
db_info = milvus_client.describe_database(db_name=db1Name)
print(db_info)
print("\nalter database properties of db1:")
milvus_client.alter_database_properties(db_name=db1Name, properties={"key": "value"})
db_info = milvus_client.describe_database(db_name=db1Name)
print(db_info)

print("\ndrop database properties of db1")
milvus_client.drop_database_properties(db_name=db1Name, property_keys=["key"])
db_info = milvus_client.describe_database(db_name=db1Name)
print(db_info)

# list database
print("\nlist databases:")
print(milvus_client.list_databases())


import json
import tempfile
from pathlib import Path
from unittest.mock import patch

import numpy as np
import pytest
from pymilvus.bulk_writer.buffer import Buffer
from pymilvus.bulk_writer.constants import DYNAMIC_FIELD_NAME, BulkFileType
from pymilvus.client.types import DataType
from pymilvus.exceptions import MilvusException
from pymilvus.orm.schema import CollectionSchema, FieldSchema


class TestBuffer:
    @pytest.fixture
    def simple_schema(self):
        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True, auto_id=False),
            FieldSchema(name="vector", dtype=DataType.FLOAT_VECTOR, dim=128),
            FieldSchema(name="text", dtype=DataType.VARCHAR, max_length=512),
        ]
        return CollectionSchema(fields=fields)

    @pytest.fixture
    def complex_schema(self):
        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True, auto_id=False),
            FieldSchema(name="bool_field", dtype=DataType.BOOL),
            FieldSchema(name="int8_field", dtype=DataType.INT8),
            FieldSchema(name="int16_field", dtype=DataType.INT16),
            FieldSchema(name="int32_field", dtype=DataType.INT32),
            FieldSchema(name="float_field", dtype=DataType.FLOAT),
            FieldSchema(name="double_field", dtype=DataType.DOUBLE),
            FieldSchema(name="json_field", dtype=DataType.JSON),
            FieldSchema(name="array_field", dtype=DataType.ARRAY, max_capacity=100, element_type=DataType.INT64),
            FieldSchema(name="vector", dtype=DataType.FLOAT_VECTOR, dim=128),
            FieldSchema(name="binary_vector", dtype=DataType.BINARY_VECTOR, dim=128),
            FieldSchema(name="sparse_vector", dtype=DataType.SPARSE_FLOAT_VECTOR),
            FieldSchema(name="float16_vector", dtype=DataType.FLOAT16_VECTOR, dim=128),
            FieldSchema(name="bfloat16_vector", dtype=DataType.BFLOAT16_VECTOR, dim=128),
            FieldSchema(name="int8_vector", dtype=DataType.INT8_VECTOR, dim=128),
        ]
        return CollectionSchema(fields=fields)

    @pytest.fixture
    def dynamic_schema(self):
        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True, auto_id=False),
            FieldSchema(name="vector", dtype=DataType.FLOAT_VECTOR, dim=128),
        ]
        return CollectionSchema(fields=fields, enable_dynamic_field=True)

    def test_buffer_init_numpy(self, simple_schema: CollectionSchema):
        buffer = Buffer(simple_schema, BulkFileType.NUMPY)
        assert buffer._file_type == BulkFileType.NUMPY
        assert len(buffer._buffer) == 3
        assert "id" in buffer._buffer
        assert "vector" in buffer._buffer
        assert "text" in buffer._buffer

    def test_buffer_init_json(self, simple_schema: CollectionSchema):
        buffer = Buffer(simple_schema, BulkFileType.JSON)
        assert buffer._file_type == BulkFileType.JSON
        assert len(buffer._buffer) == 3

    def test_buffer_init_parquet(self, simple_schema: CollectionSchema):
        buffer = Buffer(simple_schema, BulkFileType.PARQUET)
        assert buffer._file_type == BulkFileType.PARQUET
        assert len(buffer._buffer) == 3

    def test_buffer_init_csv(self, simple_schema: CollectionSchema):
        buffer = Buffer(simple_schema, BulkFileType.CSV)
        assert buffer._file_type == BulkFileType.CSV
        assert len(buffer._buffer) == 3

    def test_buffer_init_with_dynamic_field(self, dynamic_schema: CollectionSchema):
        buffer = Buffer(dynamic_schema, BulkFileType.JSON)
        assert DYNAMIC_FIELD_NAME in buffer._buffer
        assert len(buffer._buffer) == 3  # id, vector, $meta

    def test_buffer_init_auto_id_field(self):
        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True, auto_id=True),
            FieldSchema(name="vector", dtype=DataType.FLOAT_VECTOR, dim=128),
        ]
        schema = CollectionSchema(fields=fields)
        buffer = Buffer(schema, BulkFileType.JSON)
        assert "id" not in buffer._buffer
        assert "vector" in buffer._buffer

    def test_buffer_init_function_output_field(self):
        # Test that fields marked with is_function_output=True are skipped in buffer
        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True),
            FieldSchema(name="vector", dtype=DataType.FLOAT_VECTOR, dim=128),
            FieldSchema(name="output", dtype=DataType.FLOAT),
        ]
        # Manually mark the output field as function output
        # (This would normally be done by CollectionSchema when functions are provided)
        fields[2].is_function_output = True

        schema = CollectionSchema(fields=fields)

        # Verify that the field is marked as function output
        output_field = None
        for field in schema.fields:
            if field.name == "output":
                output_field = field
                break

        assert output_field is not None
        assert output_field.is_function_output is True

        # Create buffer and verify function output field is NOT included
        buffer = Buffer(schema, BulkFileType.JSON)
        assert "output" not in buffer._buffer
        assert "id" in buffer._buffer
        assert "vector" in buffer._buffer

    def test_buffer_init_only_auto_id_field(self):
        # Test with only an auto_id primary field (which gets skipped in buffer)
        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True, auto_id=True),
        ]
        schema = CollectionSchema(fields=fields)
        # Buffer should raise exception because no fields remain after filtering
        with pytest.raises(MilvusException, match="fields list is empty"):
            Buffer(schema, BulkFileType.JSON)

    def test_append_row_simple(self, simple_schema: CollectionSchema):
        buffer = Buffer(simple_schema, BulkFileType.JSON)
        row = {
            "id": 1,
            "vector": [1.0] * 128,
            "text": "test text"
        }
        buffer.append_row(row)
        assert buffer.row_count == 1
        assert buffer._buffer["id"][0] == 1
        assert buffer._buffer["vector"][0] == [1.0] * 128
        assert buffer._buffer["text"][0] == "test text"

    def test_append_row_with_numpy(self, simple_schema: CollectionSchema):
        buffer = Buffer(simple_schema, BulkFileType.JSON)
        row = {
            "id": np.int64(1),
            "vector": np.array([1.0] * 128, dtype=np.float32),
            "text": "test text"
        }
        buffer.append_row(row)
        assert buffer.row_count == 1
        assert buffer._buffer["id"][0] == 1
        assert len(buffer._buffer["vector"][0]) == 128

    def test_append_row_dynamic_field(self, dynamic_schema: CollectionSchema):
        buffer = Buffer(dynamic_schema, BulkFileType.JSON)
        row = {
            "id": 1,
            "vector": [1.0] * 128,
            "extra_field": "extra_value",
            "another_field": 123
        }
        buffer.append_row(row)
        assert buffer.row_count == 1
        assert buffer._buffer[DYNAMIC_FIELD_NAME][0]["extra_field"] == "extra_value"
        assert buffer._buffer[DYNAMIC_FIELD_NAME][0]["another_field"] == 123

    def test_append_row_dynamic_field_dict(self, dynamic_schema: CollectionSchema):
        buffer = Buffer(dynamic_schema, BulkFileType.JSON)
        row = {
            "id": 1,
            "vector": [1.0] * 128,
            DYNAMIC_FIELD_NAME: {"field1": "value1", "field2": 2}
        }
        buffer.append_row(row)
        assert buffer.row_count == 1
        assert buffer._buffer[DYNAMIC_FIELD_NAME][0]["field1"] == "value1"
        assert buffer._buffer[DYNAMIC_FIELD_NAME][0]["field2"] == 2

    def test_append_row_invalid_dynamic_field(self, dynamic_schema: CollectionSchema):
        buffer = Buffer(dynamic_schema, BulkFileType.JSON)
        row = {
            "id": 1,
            "vector": [1.0] * 128,
            DYNAMIC_FIELD_NAME: "not_a_dict"
        }
        with pytest.raises(MilvusException):
            buffer.append_row(row)

    def test_persist_npy(self, simple_schema):
        with tempfile.TemporaryDirectory() as temp_dir:
            buffer = Buffer(simple_schema, BulkFileType.NUMPY)
            buffer.append_row({
                "id": 1,
                "vector": [1.0] * 128,
                "text": "test"
            })

            files = buffer.persist(temp_dir)
            assert len(files) == 3
            # Check files were created
            for f in files:
                pf = Path(f)
                assert pf.exists()
                assert pf.suffix == '.npy'

    def test_persist_json_rows(self, simple_schema):
        with tempfile.TemporaryDirectory() as temp_dir:
            buffer = Buffer(simple_schema, BulkFileType.JSON)
            buffer.append_row({
                "id": 1,
                "vector": [1.0] * 128,
                "text": "test"
            })

            files = buffer.persist(temp_dir)
            assert len(files) == 1
            assert files[0].endswith(".json")
            # Verify file was created and contains correct data
            file = Path(files[0])
            assert file.exists()
            with file.open('r') as f:
                data = json.load(f)
                assert 'rows' in data
                assert len(data['rows']) == 1
                assert data['rows'][0]['id'] == 1

    @patch('pandas.DataFrame.to_parquet')
    def test_persist_parquet(self, mock_to_parquet, simple_schema):
        buffer = Buffer(simple_schema, BulkFileType.PARQUET)
        buffer.append_row({
            "id": 1,
            "vector": [1.0] * 128,
            "text": "test"
        })

        files = buffer.persist("/tmp/test")
        assert len(files) == 1
        assert files[0].endswith(".parquet")
        assert mock_to_parquet.called

    @patch('pandas.DataFrame.to_csv')
    def test_persist_csv(self, mock_to_csv, simple_schema):
        buffer = Buffer(simple_schema, BulkFileType.CSV)
        buffer.append_row({
            "id": 1,
            "vector": [1.0] * 128,
            "text": "test"
        })

        with patch.object(buffer, '_persist_csv', return_value=["/tmp/test.csv"]):
            files = buffer.persist("/tmp/test")
            assert len(files) == 1
            assert files[0].endswith(".csv")

    def test_persist_unsupported_type(self, simple_schema):
        buffer = Buffer(simple_schema, BulkFileType.JSON)
        buffer._file_type = 999  # Invalid type
        buffer.append_row({"id": 1, "vector": [1.0] * 128, "text": "test"})

        with pytest.raises(MilvusException):
            buffer.persist("/tmp/test")

    def test_persist_mismatched_row_count(self, simple_schema):
        buffer = Buffer(simple_schema, BulkFileType.JSON)
        buffer._buffer["id"].append(1)
        buffer._buffer["vector"].append([1.0] * 128)
        # Missing text field value

        with pytest.raises(MilvusException):
            buffer.persist("/tmp/test")

    def test_row_count_empty(self, simple_schema):
        buffer = Buffer(simple_schema, BulkFileType.JSON)
        assert buffer.row_count == 0

    def test_row_count_multiple(self, simple_schema):
        buffer = Buffer(simple_schema, BulkFileType.JSON)
        for i in range(5):
            buffer.append_row({
                "id": i,
                "vector": [1.0] * 128,
                "text": f"test {i}"
            })
        assert buffer.row_count == 5

    def test_complex_data_types(self, complex_schema):
        buffer = Buffer(complex_schema, BulkFileType.JSON)
        row = {
            "id": 1,
            "bool_field": True,
            "int8_field": 127,
            "int16_field": 32767,
            "int32_field": 2147483647,
            "float_field": 3.14,
            "double_field": 2.718281828,
            "json_field": {"key": "value"},
            "array_field": [1, 2, 3, 4, 5],
            "vector": [1.0] * 128,
            "binary_vector": [1] * 128,
            "sparse_vector": {1: 0.5, 10: 0.3},
            "float16_vector": bytes([1] * 32),
            "bfloat16_vector": bytes([1] * 32),
            "int8_vector": [1] * 128,
        }
        buffer.append_row(row)
        assert buffer.row_count == 1

    @pytest.mark.xfail(reason='numpy.dtype("bfloat16") is not supported')
    def test_persist_json_with_special_vectors(self, complex_schema):
        with tempfile.TemporaryDirectory() as temp_dir:
            buffer = Buffer(complex_schema, BulkFileType.JSON)
            row = {
                "id": 1,
                "bool_field": True,
                "int8_field": 127,
                "int16_field": 32767,
                "int32_field": 2147483647,
                "float_field": 3.14,
                "double_field": 2.718281828,
                "json_field": {"key": "value"},
                "array_field": [1, 2, 3],
                "vector": [1.0] * 128,
                "binary_vector": [1] * 128,
                "sparse_vector": {1: 0.5},
                "float16_vector": bytes([1] * 32),
                "bfloat16_vector": bytes([1] * 32),
                "int8_vector": [1] * 128,
            }
            buffer.append_row(row)

            files = buffer.persist(temp_dir)
            assert len(files) == 1
            assert Path(files[0]).exists()

    def test_persist_with_kwargs(self, simple_schema):
        buffer = Buffer(simple_schema, BulkFileType.PARQUET)
        buffer.append_row({
            "id": 1,
            "vector": [1.0] * 128,
            "text": "test"
        })

        with patch('pandas.DataFrame.to_parquet') as mock_to_parquet:
            files = buffer.persist(
                "/tmp/test",
                buffer_size=1024,
                buffer_row_count=1,
                row_group_bytes=32 * 1024 * 1024
            )
            assert len(files) == 1
            mock_to_parquet.assert_called_once()

    def test_raw_obj_conversion(self, simple_schema):
        buffer = Buffer(simple_schema, BulkFileType.JSON)

        # Test numpy array conversion
        arr = np.array([1, 2, 3])
        result = buffer._raw_obj(arr)
        assert result == [1, 2, 3]

        # Test numpy scalar conversion
        scalar = np.int64(42)
        result = buffer._raw_obj(scalar)
        assert result == 42

        # Test regular object
        obj = {"key": "value"}
        result = buffer._raw_obj(obj)
        assert result == {"key": "value"}


class TestBufferExtended:
    """Extended tests to cover additional edge cases and error paths in buffer.py"""

    @pytest.fixture
    def schema_with_array(self):
        """Schema with array field for testing array error handling"""
        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True),
            FieldSchema(name="array_field", dtype=DataType.ARRAY, max_capacity=100, element_type=DataType.INT64),
        ]
        return CollectionSchema(fields=fields)

    @pytest.fixture
    def schema_with_sparse(self):
        """Schema with sparse vector field"""
        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True),
            FieldSchema(name="sparse_vector", dtype=DataType.SPARSE_FLOAT_VECTOR),
        ]
        return CollectionSchema(fields=fields)

    def test_persist_npy_with_array_field_error(self, schema_with_array):
        """Test that array field raises error in numpy format"""
        buffer = Buffer(schema_with_array, BulkFileType.NUMPY)
        buffer.append_row({
            "id": 1,
            "array_field": [1, 2, 3, 4, 5]
        })

        with tempfile.TemporaryDirectory() as temp_dir, pytest.raises(MilvusException, match="doesn't support parsing array type"):
            buffer.persist(temp_dir)

    def test_persist_npy_with_sparse_vector_error(self, schema_with_sparse):
        """Test that sparse vector field raises error in numpy format"""
        buffer = Buffer(schema_with_sparse, BulkFileType.NUMPY)
        buffer.append_row({
            "id": 1,
            "sparse_vector": {1: 0.5, 10: 0.3}
        })

        with tempfile.TemporaryDirectory() as temp_dir, pytest.raises(MilvusException, match="SPARSE_FLOAT_VECTOR"):
            # The error happens because SPARSE_FLOAT_VECTOR is not in NUMPY_TYPE_CREATOR
            # This causes a KeyError which is caught and re-raised as MilvusException
            buffer.persist(temp_dir)

    def test_persist_npy_with_json_field(self):
        """Test persisting JSON field as numpy"""
        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True),
            FieldSchema(name="json_field", dtype=DataType.JSON),
        ]
        schema = CollectionSchema(fields=fields)
        buffer = Buffer(schema, BulkFileType.NUMPY)
        buffer.append_row({
            "id": 1,
            "json_field": {"key": "value", "nested": {"data": 123}}
        })

        with tempfile.TemporaryDirectory() as temp_dir:
            files = buffer.persist(temp_dir)
            assert len(files) == 2
            # Verify JSON field was persisted
            json_file = next(f for f in files if "json_field" in f)
            assert Path(json_file).exists()

    def test_persist_npy_with_float16_vectors(self):
        """Test persisting float16 and bfloat16 vectors as numpy"""
        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True),
            FieldSchema(name="float16_vector", dtype=DataType.FLOAT16_VECTOR, dim=4),
            FieldSchema(name="bfloat16_vector", dtype=DataType.BFLOAT16_VECTOR, dim=4),
        ]
        schema = CollectionSchema(fields=fields)
        buffer = Buffer(schema, BulkFileType.NUMPY)

        # Add data with bytes for float16/bfloat16 vectors
        float16_data = np.array([1.0, 2.0, 3.0, 4.0], dtype=np.float16).tobytes()
        bfloat16_data = bytes([1, 2, 3, 4, 5, 6, 7, 8])  # 4 bfloat16 values = 8 bytes

        buffer.append_row({
            "id": 1,
            "float16_vector": float16_data,
            "bfloat16_vector": bfloat16_data,
        })

        with tempfile.TemporaryDirectory() as temp_dir:
            files = buffer.persist(temp_dir)
            assert len(files) == 3
            for f in files:
                assert Path(f).exists()

    def test_persist_npy_with_none_values(self):
        """Test persisting None values in numpy format"""
        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True),
            FieldSchema(name="nullable_field", dtype=DataType.INT64),
        ]
        schema = CollectionSchema(fields=fields)
        buffer = Buffer(schema, BulkFileType.NUMPY)
        buffer.append_row({"id": 1, "nullable_field": None})
        buffer.append_row({"id": 2, "nullable_field": 100})

        with tempfile.TemporaryDirectory() as temp_dir:
            files = buffer.persist(temp_dir)
            assert len(files) == 2

    @patch('numpy.save')
    def test_persist_npy_exception_handling(self, mock_save):
        """Test exception handling in numpy persist"""
        mock_save.side_effect = Exception("Save failed")

        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True),
        ]
        schema = CollectionSchema(fields=fields)
        buffer = Buffer(schema, BulkFileType.NUMPY)
        buffer.append_row({"id": 1})

        with tempfile.TemporaryDirectory() as temp_dir, pytest.raises(MilvusException, match="Failed to persist file"):
            buffer.persist(temp_dir)

    def test_persist_npy_cleanup_on_partial_failure(self):
        """Test that files are cleaned up if not all fields persist successfully"""
        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True),
            FieldSchema(name="vector", dtype=DataType.FLOAT_VECTOR, dim=2),
        ]
        schema = CollectionSchema(fields=fields)
        buffer = Buffer(schema, BulkFileType.NUMPY)
        buffer.append_row({"id": 1, "vector": [1.0, 2.0]})

        with tempfile.TemporaryDirectory() as temp_dir, patch('numpy.save') as mock_save:
            # Make the second save fail
            mock_save.side_effect = [None, Exception("Second save failed")]

            with pytest.raises(MilvusException, match="Failed to persist file"):
                buffer.persist(temp_dir)

    def test_persist_json_with_float16_vectors(self):
        """Test JSON persist with float16 vectors only"""
        # Skip bfloat16 as numpy doesn't have native support
        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True),
            FieldSchema(name="float16_vector", dtype=DataType.FLOAT16_VECTOR, dim=4),
        ]
        schema = CollectionSchema(fields=fields)
        buffer = Buffer(schema, BulkFileType.JSON)

        float16_data = np.array([1.0, 2.0, 3.0, 4.0], dtype=np.float16).tobytes()

        buffer.append_row({
            "id": 1,
            "float16_vector": float16_data,
        })

        with tempfile.TemporaryDirectory() as temp_dir:
            files = buffer.persist(temp_dir)
            assert len(files) == 1

            # Verify the JSON content
            with Path(files[0]).open('r') as f:
                data = json.load(f)
                assert 'rows' in data
                assert len(data['rows']) == 1
                assert isinstance(data['rows'][0]['float16_vector'], list)

    @patch('pathlib.Path.open')
    def test_persist_json_exception_handling(self, mock_open):
        """Test exception handling in JSON persist"""
        mock_open.side_effect = Exception("Failed to open file")

        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True),
        ]
        schema = CollectionSchema(fields=fields)
        buffer = Buffer(schema, BulkFileType.JSON)
        buffer.append_row({"id": 1})

        with pytest.raises(MilvusException, match="Failed to persist file"):
            buffer.persist("/tmp/test")

    def test_persist_parquet_with_json_and_sparse(self):
        """Test Parquet persist with JSON and sparse vector fields"""
        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True),
            FieldSchema(name="json_field", dtype=DataType.JSON),
            FieldSchema(name="sparse_vector", dtype=DataType.SPARSE_FLOAT_VECTOR),
        ]
        schema = CollectionSchema(fields=fields)
        buffer = Buffer(schema, BulkFileType.PARQUET)

        buffer.append_row({
            "id": 1,
            "json_field": {"key": "value"},
            "sparse_vector": {1: 0.5, 10: 0.3}
        })

        with tempfile.TemporaryDirectory() as temp_dir, patch('pandas.DataFrame.to_parquet') as mock_to_parquet:
            # Pass buffer_size and buffer_row_count to avoid UnboundLocalError
            files = buffer.persist(temp_dir, buffer_size=1024, buffer_row_count=1)
            assert len(files) == 1
            mock_to_parquet.assert_called_once()

    def test_persist_parquet_with_float16_vectors(self):
        """Test Parquet persist with float16 vectors only"""
        # Skip bfloat16 as it has issues with numpy dtype
        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True),
            FieldSchema(name="float16_vector", dtype=DataType.FLOAT16_VECTOR, dim=4),
        ]
        schema = CollectionSchema(fields=fields)
        buffer = Buffer(schema, BulkFileType.PARQUET)

        float16_data = np.array([1.0, 2.0, 3.0, 4.0], dtype=np.float16).tobytes()

        buffer.append_row({
            "id": 1,
            "float16_vector": float16_data,
        })

        with tempfile.TemporaryDirectory() as temp_dir, patch('pandas.DataFrame.to_parquet') as mock_to_parquet:
            # Pass buffer_size and buffer_row_count to avoid UnboundLocalError
            files = buffer.persist(temp_dir, buffer_size=1024, buffer_row_count=1)
            assert len(files) == 1
            mock_to_parquet.assert_called_once()

    def test_persist_parquet_with_array_field(self):
        """Test Parquet persist with array field"""
        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True),
            FieldSchema(name="array_field", dtype=DataType.ARRAY, max_capacity=100, element_type=DataType.INT64),
        ]
        schema = CollectionSchema(fields=fields)
        buffer = Buffer(schema, BulkFileType.PARQUET)

        buffer.append_row({
            "id": 1,
            "array_field": [1, 2, 3, 4, 5]
        })
        buffer.append_row({
            "id": 2,
            "array_field": None  # Test None value
        })

        with tempfile.TemporaryDirectory() as temp_dir, patch('pandas.DataFrame.to_parquet') as mock_to_parquet:
            # Pass buffer_size and buffer_row_count to avoid UnboundLocalError
            files = buffer.persist(temp_dir, buffer_size=1024, buffer_row_count=2)
            assert len(files) == 1
            mock_to_parquet.assert_called_once()

    def test_persist_parquet_with_unknown_dtype(self):
        """Test Parquet persist with fields having standard dtypes"""
        # Using standard field types that work with parquet
        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True),
            FieldSchema(name="field1", dtype=DataType.VARCHAR, max_length=100),
            FieldSchema(name="field2", dtype=DataType.JSON),
        ]
        schema = CollectionSchema(fields=fields)
        buffer = Buffer(schema, BulkFileType.PARQUET)

        buffer.append_row({
            "id": 1,
            "field1": "test_value",
            "field2": {"key": "value"}
        })

        with tempfile.TemporaryDirectory() as temp_dir, patch('pandas.DataFrame.to_parquet') as mock_to_parquet:
            # Pass buffer_size and buffer_row_count to avoid UnboundLocalError
            files = buffer.persist(temp_dir, buffer_size=1024, buffer_row_count=1)
            assert len(files) == 1
            mock_to_parquet.assert_called_once()

    def test_persist_csv_with_all_field_types(self):
        """Test CSV persist with common field types"""
        # Skip bfloat16 due to numpy dtype issues
        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True),
            FieldSchema(name="bool_field", dtype=DataType.BOOL),
            FieldSchema(name="int8_field", dtype=DataType.INT8),
            FieldSchema(name="int16_field", dtype=DataType.INT16),
            FieldSchema(name="int32_field", dtype=DataType.INT32),
            FieldSchema(name="float_field", dtype=DataType.FLOAT),
            FieldSchema(name="double_field", dtype=DataType.DOUBLE),
            FieldSchema(name="varchar_field", dtype=DataType.VARCHAR, max_length=512),
            FieldSchema(name="json_field", dtype=DataType.JSON),
            FieldSchema(name="array_field", dtype=DataType.ARRAY, max_capacity=100, element_type=DataType.INT64),
            FieldSchema(name="float_vector", dtype=DataType.FLOAT_VECTOR, dim=128),
            FieldSchema(name="binary_vector", dtype=DataType.BINARY_VECTOR, dim=128),
            FieldSchema(name="sparse_vector", dtype=DataType.SPARSE_FLOAT_VECTOR),
            FieldSchema(name="float16_vector", dtype=DataType.FLOAT16_VECTOR, dim=128),
            FieldSchema(name="int8_vector", dtype=DataType.INT8_VECTOR, dim=128),
        ]
        schema = CollectionSchema(fields=fields)
        buffer = Buffer(schema, BulkFileType.CSV)

        row = {
            "id": 1,
            "bool_field": True,
            "int8_field": 127,
            "int16_field": 32767,
            "int32_field": 2147483647,
            "float_field": 3.14,
            "double_field": 2.718281828,
            "varchar_field": "test string",
            "json_field": {"key": "value"},
            "array_field": [1, 2, 3],
            "float_vector": [1.0] * 128,
            "binary_vector": [1] * 128,
            "sparse_vector": {1: 0.5},
            "float16_vector": np.array([1.0] * 128, dtype=np.float16).tobytes(),
            "int8_vector": [1] * 128,
        }
        buffer.append_row(row)

        # Add row with None values
        row_with_none = row.copy()
        row_with_none["array_field"] = None
        buffer.append_row(row_with_none)

        with tempfile.TemporaryDirectory() as temp_dir, patch('pandas.DataFrame.to_csv') as mock_to_csv:
            files = buffer.persist(temp_dir)
            assert len(files) == 1
            mock_to_csv.assert_called_once()

    @patch('pandas.DataFrame.to_csv')
    def test_persist_csv_exception_handling(self, mock_to_csv):
        """Test exception handling in CSV persist"""
        mock_to_csv.side_effect = Exception("CSV write failed")

        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True),
        ]
        schema = CollectionSchema(fields=fields)
        buffer = Buffer(schema, BulkFileType.CSV)
        buffer.append_row({"id": 1})

        with pytest.raises(MilvusException, match="Failed to persist file"):
            buffer.persist("/tmp/test")

    def test_persist_csv_with_custom_config(self):
        """Test CSV persist with custom separator and null key"""
        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True),
            FieldSchema(name="nullable_field", dtype=DataType.INT64),
        ]
        schema = CollectionSchema(fields=fields)

        # Create buffer with custom CSV config
        config = {"sep": "|", "nullkey": "NULL"}
        buffer = Buffer(schema, BulkFileType.CSV, config=config)

        buffer.append_row({"id": 1, "nullable_field": None})
        buffer.append_row({"id": 2, "nullable_field": 100})

        with tempfile.TemporaryDirectory() as temp_dir, patch('pandas.DataFrame.to_csv') as mock_to_csv:
            files = buffer.persist(temp_dir)
            assert len(files) == 1

            # Verify custom config was used
            call_kwargs = mock_to_csv.call_args[1]
            assert call_kwargs.get('sep') == '|'
            assert call_kwargs.get('na_rep') == 'NULL'

    def test_get_field_schema(self):
        """Test accessing fields from buffer"""
        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True),
            FieldSchema(name="vector", dtype=DataType.FLOAT_VECTOR, dim=128),
        ]
        schema = CollectionSchema(fields=fields)
        buffer = Buffer(schema, BulkFileType.JSON)

        # Test that fields are stored in the buffer
        assert "id" in buffer._fields
        assert "vector" in buffer._fields
        assert buffer._fields["id"].name == "id"
        assert buffer._fields["vector"].name == "vector"

        # Test that non-existent field is not in buffer
        assert "non_existent" not in buffer._fields

    def test_throw_method(self):
        """Test the _throw method raises MilvusException"""
        fields = [
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True),
        ]
        schema = CollectionSchema(fields=fields)
        buffer = Buffer(schema, BulkFileType.JSON)

        with pytest.raises(MilvusException, match="Test error message"):
            buffer._throw("Test error message")



import json
import struct
from typing import List
from unittest.mock import patch

import numpy as np
import pytest
from pymilvus.client import entity_helper
from pymilvus.client.entity_helper import (
    convert_to_str_array, entity_to_array_arr,
    entity_to_field_data,
    entity_to_str_arr,
    entity_type_to_dtype,
    extract_array_row_data,
    extract_dynamic_field_from_result,
    get_max_len_of_var_char,
    pack_field_value_to_field_data,
    sparse_proto_to_rows,
    sparse_rows_to_proto,
)
from pymilvus.client.entity_helper import extract_row_data_from_fields_data_v2 as convert_to_entity
from pymilvus.client.types import DataType
from pymilvus.exceptions import ParamError
from pymilvus.grpc_gen import schema_pb2

# Import additional functions for testing
from pymilvus.grpc_gen import schema_pb2 as schema_types
from pymilvus.settings import Config
from scipy.sparse import csr_matrix
from pymilvus.exceptions import DataNotMatchException


class TestEntityHelperSparse:
    """Test entity_helper module functions"""

    @pytest.mark.parametrize("valid_sparse_matrix", [
        [{0: 1.0, 5: 2.5, 10: 3.0}], # list of one dict
        [{0: 1.0, 5: 2.5}, {10: 3.0, 15: 4.0}], # list of dicts
        [{}, {10: 3.0, 15: 4.0}], # list of dicts partial empty is allowed
        [[(1, 0.5), (10, 0.3)], [(2, 0.7), (20, 0.1)]], # list of list
        [[("1", "0.5"), (10, 0.3)]], # str representation of int
        csr_matrix(([1, 2, 3], [0, 2, 3], [0, 2, 3, 3]), shape=(3, 4)), # scipy sparse matrix
        [csr_matrix([[1, 0, 2]]), csr_matrix([[0, 0, 3]])], # list of scipy sparse matrices
    ])
    def test_entity_is_sparse_matrix(self, valid_sparse_matrix: list):
        assert entity_helper.entity_is_sparse_matrix(valid_sparse_matrix) is True

    @pytest.mark.parametrize("not_sparse_matrix", [
        [{"a": 1.0, "b": 2.0}], # invalid dict for non-numeric keys
        [], # empty
        [{0: 1.0}, "not a dict", {5: 2.0}], # mixed lists
        None,
        123,
        "string",
        [1, 2, 3],
        [[1, 2, 3]],
        [[(1, 0.5, 0.2)]],
        [[(1, "invalid")]],
        [csr_matrix([[1, 0], [0, 1]])], # list of multi-row is not sparse
    ])
    def test_entity_isnot_sparse_matrix(self, not_sparse_matrix: any):
        assert entity_helper.entity_is_sparse_matrix(not_sparse_matrix) is False

    def test_get_input_num_rows_list(self):
        """Test getting number of rows from list input"""
        # Regular list
        data = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        assert entity_helper.get_input_num_rows(data) == 3
        assert entity_helper.get_input_num_rows([1, 2, 3]) == 3
        assert entity_helper.get_input_num_rows({"a": 1, "b": 2}) == 2

        data = [[1, 2, 3]]
        assert entity_helper.get_input_num_rows(data) == 1

        data = []
        assert entity_helper.get_input_num_rows(data) == 0

        matrix = csr_matrix([[1, 0], [0, 1], [1, 1]])
        assert entity_helper.get_input_num_rows(matrix) == 3

        sparse_list = [
            {0: 1.0},
            {5: 2.5},
            {10: 3.0}
        ]
        assert entity_helper.get_input_num_rows(sparse_list) == 3

        data = np.array([[1, 2, 3], [4, 5, 6]])
        assert entity_helper.get_input_num_rows(data) == 2

    @pytest.mark.parametrize("sparse_list", [
        [{0: 1.0, 2: 2.0}, {2: 3.0}],
        csr_matrix([[1, 0, 2], [0, 0, 3]])
    ])
    def test_sparse_rows_to_proto_dict(self, sparse_list: any):
        """Test converting sparse rows to protobuf format"""
        proto = entity_helper.sparse_rows_to_proto(sparse_list)
        assert isinstance(proto, schema_pb2.SparseFloatArray)
        assert len(proto.contents) == 2
        assert proto.dim == 3

    def test_sparse_proto_to_rows(self):
        """Test converting protobuf sparse vectors back to rows"""
        # Create a mock sparse proto
        proto = schema_pb2.SparseFloatArray(dim=100)

        # Add some sparse vectors in binary format
        # Format: pairs of (uint32 index, float32 value)
        vec1_data = b''
        for idx, val in [(0, 1.0), (5, 2.5), (10, 3.0)]:
            vec1_data += struct.pack('I', idx) + struct.pack('f', val)
        proto.contents.append(vec1_data)

        vec2_data = b''
        for idx, val in [(15, 4.0), (20, 5.0)]:
            vec2_data += struct.pack('I', idx) + struct.pack('f', val)
        proto.contents.append(vec2_data)

        # Convert back to rows
        rows = entity_helper.sparse_proto_to_rows(proto, 0, 2)
        assert len(rows) == 2
        assert rows[0] == {0: 1.0, 5: 2.5, 10: 3.0}
        assert rows[1] == {15: 4.0, 20: 5.0}

    def test_sparse_proto_to_rows_with_range(self):
        """Test converting specific range of sparse vectors"""
        proto = schema_pb2.SparseFloatArray(dim=100)

        # Add multiple sparse vectors in binary format
        for i in range(5):
            vec_data = struct.pack('I', i) + struct.pack('f', float(i))
            proto.contents.append(vec_data)

        # Get middle range
        rows = entity_helper.sparse_proto_to_rows(proto, 1, 4)
        assert len(rows) == 3
        assert rows[0] == {1: 1.0}
        assert rows[1] == {2: 2.0}
        assert rows[2] == {3: 3.0}

    def test_convert_to_json_nested(self):
        """Test JSON conversion with nested structures"""
        obj = {
            "level1": {
                "level2": {
                    "array": np.array([[1, 2], [3, 4]]),
                    "list": [np.int32(1), np.float64(2.5)]
                }
            }
        }
        result = entity_helper.convert_to_json(obj)
        parsed = json.loads(result)
        assert parsed["level1"]["level2"]["array"] == [[1, 2], [3, 4]]

    @pytest.mark.parametrize("data", [
        {"key": "value", "number": 42}, # dict
        {"outer": {"inner": "value"}}, # nested dict
        [1, 2, 3, "four"], # list
        [{"a": 1}, {"b": 2}], # list of dict
        None,
        pytest.param({"array": np.array([1, 2, 3])}, marks=pytest.mark.xfail(reason="fix me")),
        { "int": np.int64(42), "float": np.float32(3.14), "bool": np.bool_(True) },
        [{"val": np.int64(10)}, {"val": np.float32(3.14)}],
    ])
    def test_convert_to_json_dict(self, data: dict):
        """Test JSON conversion for dict input"""
        result = entity_helper.convert_to_json(data)
        assert isinstance(result, bytes)
        assert json.loads(result.decode()) == data

    @pytest.mark.parametrize("json_string,expected", [
        ('{"key": "value", "number": 42}', {"key": "value", "number": 42}),
        ('{"nested": {"inner": "value"}}', {"nested": {"inner": "value"}}),
        ('[1, 2, 3, "four"]', [1, 2, 3, "four"]),
        ('{"name": "Alice", "age": 30}', {"name": "Alice", "age": 30}),
        ('null', None),
        ('true', True),
        ('false', False),
        ('123', 123),
        ('"simple string"', "simple string"),
    ])
    def test_convert_to_json_string_valid(self, json_string: str, expected):
        """Test JSON conversion for valid JSON string input"""
        result = entity_helper.convert_to_json(json_string)
        assert isinstance(result, bytes)
        # Verify the result is valid JSON
        parsed = json.loads(result.decode())
        assert parsed == expected

    def test_convert_to_json_from_json_dumps(self):
        """Test JSON conversion from json.dumps() output"""
        original_dict = {"key": "value", "count": 100, "nested": {"inner": "data"}}
        json_string = json.dumps(original_dict)
        
        result = entity_helper.convert_to_json(json_string)
        assert isinstance(result, bytes)
        parsed = json.loads(result.decode())
        assert parsed == original_dict

    @pytest.mark.parametrize("invalid_json_string", [
        "not a json string",
        '{"invalid": }',
        '{"key": "value"',  # missing closing brace
        "{'key': 'value'}",  # single quotes not valid in JSON
        "{key: value}",  # unquoted keys
        "undefined",
        "{,}",
    ])
    def test_convert_to_json_string_invalid(self, invalid_json_string: str):
        """Test JSON conversion rejects invalid JSON strings"""
        
        with pytest.raises(DataNotMatchException) as exc_info:
            entity_helper.convert_to_json(invalid_json_string)
        
        # Verify error message contains the invalid JSON string
        error_message = str(exc_info.value)
        assert "Invalid JSON string" in error_message
        # Verify the original input string is in the error message
        assert invalid_json_string in error_message or invalid_json_string[:50] in error_message

    def test_convert_to_json_string_with_non_string_keys(self):
        """Test JSON conversion rejects JSON strings with non-string keys in dict"""
        
        # This is actually not possible in standard JSON, as JSON object keys are always strings
        # But we can test that dict validation still works
        invalid_dict = {1: "value", 2: "another"}
        
        with pytest.raises(DataNotMatchException) as exc_info:
            entity_helper.convert_to_json(invalid_dict)
        
        error_message = str(exc_info.value)
        assert "JSON" in error_message

    def test_convert_to_json_long_invalid_string_truncated(self):
        """Test that long invalid JSON strings are truncated in error messages"""
        
        # Create a long invalid JSON string
        long_invalid_json = "invalid json " * 50  # > 200 characters
        
        with pytest.raises(DataNotMatchException) as exc_info:
            entity_helper.convert_to_json(long_invalid_json)
        
        error_message = str(exc_info.value)
        assert "Invalid JSON string" in error_message
        # Should contain truncated version with "..."
        assert "..." in error_message

    def test_pack_field_value_to_field_data(self):
        """Test packing field values into field data protobuf"""
        # Test with scalar field
        field_data = schema_pb2.FieldData()
        field_data.type = DataType.INT64
        field_info = {"name": "test_field"}

        value = 42
        entity_helper.pack_field_value_to_field_data(value, field_data, field_info)

        assert len(field_data.scalars.long_data.data) == 1
        assert field_data.scalars.long_data.data[0] == value

    def test_pack_field_value_to_field_data_vectors(self):
        """Test packing vector field values"""
        field_data = schema_pb2.FieldData()
        field_data.type = DataType.FLOAT_VECTOR
        field_info = {"name": "vector_field"}

        value = [1.0, 2.0, 3.0, 4.0]

        entity_helper.pack_field_value_to_field_data(value, field_data, field_info)

        assert field_data.vectors.dim == 4
        assert list(field_data.vectors.float_vector.data) == value

    def test_extract_array_row_data(self):
        """Test extracting array data from protobuf"""
        # Create field data with array
        field_data = schema_pb2.FieldData()
        field_data.scalars.array_data.element_type = DataType.INT64

        # Add array data for index 0
        scalar_field = schema_pb2.ScalarField()
        scalar_field.long_data.data.extend([0, 1, 2])
        field_data.scalars.array_data.data.append(scalar_field)

        result = entity_helper.extract_array_row_data(field_data, 0)
        assert result == [0, 1, 2]

    def test_extract_array_row_data_string(self):
        """Test extracting string array data"""
        # Create field data with array
        field_data = schema_pb2.FieldData()
        field_data.scalars.array_data.element_type = DataType.VARCHAR

        # Add array data for index 0
        scalar_field = schema_pb2.ScalarField()
        scalar_field.string_data.data.extend(["str_0", "str_1"])
        field_data.scalars.array_data.data.append(scalar_field)

        result = entity_helper.extract_array_row_data(field_data, 0)
        assert result == ["str_0", "str_1"]

    def test_extract_array_row_data_bool(self):
        """Test extracting boolean array data"""
        # Create field data with array
        field_data = schema_pb2.FieldData()
        field_data.scalars.array_data.element_type = DataType.BOOL

        # Add array data for index 0
        scalar_field = schema_pb2.ScalarField()
        scalar_field.bool_data.data.extend([True, False, True])
        field_data.scalars.array_data.data.append(scalar_field)

        result = entity_helper.extract_array_row_data(field_data, 0)
        assert result == [True, False, True]

    def test_extract_array_row_data_float(self):
        """Test extracting float array data"""
        # Create field data with array
        field_data = schema_pb2.FieldData()
        field_data.scalars.array_data.element_type = DataType.FLOAT

        # Add array data for index 0
        scalar_field = schema_pb2.ScalarField()
        scalar_field.float_data.data.extend([1.1, 2.2, 3.3])
        field_data.scalars.array_data.data.append(scalar_field)

        result = entity_helper.extract_array_row_data(field_data, 0)
        assert result == pytest.approx([1.1, 2.2, 3.3])

    def test_extract_array_row_data_double(self):
        """Test extracting double array data"""
        # Create field data with array
        field_data = schema_pb2.FieldData()
        field_data.scalars.array_data.element_type = DataType.DOUBLE

        # Add array data for index 0
        scalar_field = schema_pb2.ScalarField()
        scalar_field.double_data.data.extend([1.11111, 2.22222, 3.33333])
        field_data.scalars.array_data.data.append(scalar_field)

        result = entity_helper.extract_array_row_data(field_data, 0)
        assert result == pytest.approx([1.11111, 2.22222, 3.33333])

    def test_extract_array_row_data_invalid_type(self):
        """Test extracting array data with invalid type"""
        # Create field data with array
        field_data = schema_pb2.FieldData()
        field_data.scalars.array_data.element_type = 999  # Invalid type

        # Add array data for index 0
        scalar_field = schema_pb2.ScalarField()
        field_data.scalars.array_data.data.append(scalar_field)

        result = entity_helper.extract_array_row_data(field_data, 0)
        assert result is None  # Returns None for unknown types


class TestEntityHelperExtended:
    def test_sparse_rows_to_proto_invalid_index(self):
        """Test error handling for invalid indices"""
        # Negative index
        with pytest.raises(ParamError, match="sparse vector index must be positive"):
            sparse_rows_to_proto([{-1: 0.5}])

        # Index too large
        with pytest.raises(ParamError, match="sparse vector index must be positive"):
            sparse_rows_to_proto([{2**32: 0.5}])

    def test_sparse_rows_to_proto_nan_value(self):
        """Test error handling for NaN values"""
        with pytest.raises(ParamError, match="sparse vector value must not be NaN"):
            sparse_rows_to_proto([{1: float('nan')}])

    def test_sparse_rows_to_proto_invalid_input(self):
        """Test error handling for invalid input"""
        with pytest.raises(ParamError, match="input must be a sparse matrix"):
            sparse_rows_to_proto("invalid")

    def test_sparse_proto_to_rows_invalid(self):
        """Test error handling for invalid proto"""
        with pytest.raises(ParamError, match="Vector must be a sparse float vector"):
            sparse_proto_to_rows("invalid")

    def test_entity_type_to_dtype(self):
        """Test converting entity type to dtype"""
        # Integer type
        assert entity_type_to_dtype(1) == 1
        assert entity_type_to_dtype(DataType.INT64) == DataType.INT64

        # We can't test string conversion without knowing exact protobuf enum names
        # Let's just test invalid type
        with pytest.raises(ParamError, match="invalid entity type"):
            entity_type_to_dtype([])

    def test_get_max_len_of_var_char(self):
        """Test getting max length of varchar field"""
        # With params
        field_info = {"params": {Config.MaxVarCharLengthKey: 100}}
        assert get_max_len_of_var_char(field_info) == 100

        # Without params - use default
        field_info = {}
        assert get_max_len_of_var_char(field_info) == Config.MaxVarCharLength

        # Partial params
        field_info = {"params": {}}
        assert get_max_len_of_var_char(field_info) == Config.MaxVarCharLength

    def test_convert_to_str_array(self):
        """Test converting to string array"""
        field_info = {"name": "test_field", "params": {Config.MaxVarCharLengthKey: 10}}

        # Valid strings
        result = convert_to_str_array(["hello", "world"], field_info)
        assert result == ["hello", "world"]

        # String exceeding max length
        with pytest.raises(ParamError, match="length of string exceeds max length"):
            convert_to_str_array(["this is too long"], field_info)

        # Non-string input
        with pytest.raises(ParamError, match="expects string input"):
            convert_to_str_array([123], field_info)

        # Without check
        result = convert_to_str_array([123, "test"], field_info, check=False)
        assert len(result) == 2

    @patch('pymilvus.client.entity_helper.Config')
    def test_convert_to_str_array_with_encoding(self, mock_config):
        """Test string array conversion with different encoding"""
        mock_config.EncodeProtocol = "latin-1"
        mock_config.MaxVarCharLengthKey = "max_length"
        mock_config.MaxVarCharLength = 65535

        field_info = {"name": "test_field"}
        strings = ["hello", "world"]
        result = convert_to_str_array(strings, field_info, check=False)
        assert len(result) == 2

    def test_entity_to_str_arr(self):
        """Test entity_to_str_arr wrapper function"""
        field_info = {"name": "test_field", "params": {Config.MaxVarCharLengthKey: 20}}
        result = entity_to_str_arr(["test1", "test2"], field_info)
        assert result == ["test1", "test2"]

    def test_extract_array_row_data(self):
        """Test extracting array row data"""
        # INT64 array
        field_data = schema_types.FieldData()
        field_data.type = DataType.ARRAY
        array_data = schema_types.ArrayArray()

        # Add INT64 data
        int_array = schema_types.LongArray()
        int_array.data.extend([1, 2, 3])
        array_data.data.append(schema_types.ScalarField(long_data=int_array))
        int_array2 = schema_types.LongArray()
        int_array2.data.extend([4, 5])
        array_data.data.append(schema_types.ScalarField(long_data=int_array2))

        array_data.element_type = DataType.INT64
        field_data.scalars.array_data.CopyFrom(array_data)

        result = extract_array_row_data(field_data, 0)
        assert result == [1, 2, 3]
        result = extract_array_row_data(field_data, 1)
        assert result == [4, 5]

    def test_extract_array_row_data_string(self):
        """Test extracting string array data"""
        field_data = schema_types.FieldData()
        field_data.type = DataType.ARRAY
        array_data = schema_types.ArrayArray()

        # Add string data
        str_array = schema_types.StringArray()
        str_array.data.extend(["hello", "world"])
        array_data.data.append(schema_types.ScalarField(string_data=str_array))

        array_data.element_type = DataType.VARCHAR
        field_data.scalars.array_data.CopyFrom(array_data)

        result = extract_array_row_data(field_data, 0)
        assert result == ["hello", "world"]

    def test_extract_array_row_data_bool(self):
        """Test extracting boolean array data"""
        field_data = schema_types.FieldData()
        field_data.type = DataType.ARRAY
        array_data = schema_types.ArrayArray()

        # Add bool data
        bool_array = schema_types.BoolArray()
        bool_array.data.extend([True, False, True])
        array_data.data.append(schema_types.ScalarField(bool_data=bool_array))

        array_data.element_type = DataType.BOOL
        field_data.scalars.array_data.CopyFrom(array_data)

        result = extract_array_row_data(field_data, 0)
        assert result == [True, False, True]

    def test_extract_array_row_data_float(self):
        """Test extracting float array data"""
        field_data = schema_types.FieldData()
        field_data.type = DataType.ARRAY
        array_data = schema_types.ArrayArray()

        # Add float data
        float_array = schema_types.FloatArray()
        float_array.data.extend([1.1, 2.2, 3.3])
        array_data.data.append(schema_types.ScalarField(float_data=float_array))

        array_data.element_type = DataType.FLOAT
        field_data.scalars.array_data.CopyFrom(array_data)

        result = extract_array_row_data(field_data, 0)
        assert len(result) == 3
        assert result[0] == pytest.approx(1.1)

    def test_extract_array_row_data_double(self):
        """Test extracting double array data"""
        field_data = schema_types.FieldData()
        field_data.type = DataType.ARRAY
        array_data = schema_types.ArrayArray()

        # Add double data
        double_array = schema_types.DoubleArray()
        double_array.data.extend([1.11111, 2.22222])
        array_data.data.append(schema_types.ScalarField(double_data=double_array))

        array_data.element_type = DataType.DOUBLE
        field_data.scalars.array_data.CopyFrom(array_data)

        result = extract_array_row_data(field_data, 0)
        assert len(result) == 2

    def test_extract_array_row_data_invalid_type(self):
        """Test error handling for invalid array element type"""
        field_data = schema_types.FieldData()
        field_data.type = DataType.ARRAY
        array_data = schema_types.ArrayArray()
        array_data.element_type = 999  # Invalid type
        # Add at least one empty element to avoid index error
        array_data.data.append(schema_types.ScalarField())
        field_data.scalars.array_data.CopyFrom(array_data)

        assert extract_array_row_data(field_data, 0) is None

    def test_entity_to_array_arr(self):
        """Test converting entity to array array"""
        field_info = {
            "name": "array_field",
            "element_type": DataType.INT64
        }

        # List of lists
        data = [[1, 2, 3], [4, 5], [6]]
        result = entity_to_array_arr(data, field_info)
        assert len(result) == 3
        assert result[0].long_data.data == [1, 2, 3]
        assert result[1].long_data.data == [4, 5]

    def test_entity_to_array_arr_string(self):
        """Test converting string arrays"""
        field_info = {
            "name": "array_field",
            "element_type": DataType.VARCHAR
        }

        data = [["hello", "world"], ["foo"]]
        result = entity_to_array_arr(data, field_info)
        assert len(result) == 2
        assert list(result[0].string_data.data) == ["hello", "world"]

    def test_entity_to_array_arr_invalid_type(self):
        """Test error handling for invalid element type"""
        field_info = {
            "name": "array_field",
            "element_type": 999
        }

        with pytest.raises(ParamError, match="Unsupported element type"):
            entity_to_array_arr([[1, 2]], field_info)

    def test_pack_field_value_to_field_data(self):
        """Test packing field values to field data"""
        # pack_field_value_to_field_data takes different parameters
        field_data = schema_types.FieldData()
        field_data.type = DataType.FLOAT_VECTOR
        field_data.field_name = "vector_field"
        field_info = {"name": "vector_field"}

        # Pack a single vector
        pack_field_value_to_field_data(
            np.array([1.0, 2.0]),
            field_data,
            field_info
        )

        # Check the result
        assert field_data.type == DataType.FLOAT_VECTOR
        assert field_data.vectors.dim == 2

    def test_pack_field_value_to_field_data_sparse(self):
        """Test packing sparse vectors"""
        field_data = schema_types.FieldData()
        field_data.type = DataType.SPARSE_FLOAT_VECTOR
        field_data.field_name = "sparse_field"
        field_info = {"name": "sparse_field"}

        # Pack a single sparse vector
        sparse_data = {1: 0.5, 10: 0.3}
        pack_field_value_to_field_data(
            sparse_data,
            field_data,
            field_info
        )

        # Check the result
        assert field_data.type == DataType.SPARSE_FLOAT_VECTOR
        assert len(field_data.vectors.sparse_float_vector.contents) == 1

    def test_pack_field_value_to_field_data_scalars(self):
        """Test packing scalar field values"""
        # Test INT64
        field_data = schema_types.FieldData()
        field_data.type = DataType.INT64
        field_data.field_name = "int_field"
        field_info = {"name": "int_field"}

        pack_field_value_to_field_data(
            42,
            field_data,
            field_info
        )

        assert field_data.type == DataType.INT64
        assert field_data.scalars.long_data.data[0] == 42

    def test_extract_field_info(self):
        """Test extracting primary field from schema"""
        # Create schema with primary field
        fields_info = [
            {"name": "id", "is_primary": True},
            {"name": "vector", "is_primary": False}
        ]

        # Test field extraction logic
        # This tests the field extraction patterns used in entity_helper
        for field in fields_info:
            if field.get("is_primary"):
                assert field["name"] == "id"

    def test_extract_dynamic_field_from_result(self):
        """Test extracting dynamic field from result"""
        # Create actual result object with fields_data and output_fields attributes
        class ActualResult:
            def __init__(self, fields_data: List, output_fields: List):
                self.fields_data = fields_data
                self.output_fields = output_fields

        # Test with dynamic field
        dynamic_field_data = schema_types.FieldData()
        dynamic_field_data.is_dynamic = True
        dynamic_field_data.field_name = "$meta"

        regular_field_data = schema_types.FieldData()
        regular_field_data.is_dynamic = False
        regular_field_data.field_name = "id"

        # Create result with dynamic field - extra_field comes before $meta
        result = ActualResult(
            fields_data=[regular_field_data, dynamic_field_data],
            output_fields=["id", "extra_field", "another_extra", "$meta"]
        )

        dynamic_field_name, dynamic_fields = extract_dynamic_field_from_result(result)
        assert dynamic_field_name == "$meta"
        # When $meta is found in output_fields, dynamic_fields gets cleared
        assert len(dynamic_fields) == 0

        # Test with no dynamic field
        result_no_dynamic = ActualResult(
            fields_data=[regular_field_data],
            output_fields=["id", "extra_field"]
        )

        dynamic_field_name, dynamic_fields = extract_dynamic_field_from_result(result_no_dynamic)
        assert dynamic_field_name is None
        assert "extra_field" in dynamic_fields
        assert "id" not in dynamic_fields

        # Test with dynamic field NOT in output_fields (dynamic_fields preserved)
        result_meta_not_in_output = ActualResult(
            fields_data=[regular_field_data, dynamic_field_data],
            output_fields=["id", "extra_field", "another_extra"]
        )

        dynamic_field_name, dynamic_fields = extract_dynamic_field_from_result(result_meta_not_in_output)
        assert dynamic_field_name == "$meta"
        assert "extra_field" in dynamic_fields
        assert "another_extra" in dynamic_fields
        assert "id" not in dynamic_fields  # id is a regular field
        assert len(dynamic_fields) == 2

    def test_data_validation(self):
        """Test data validation patterns"""
        # Test length validation
        data1 = [1, 2, 3]
        data2 = [4, 5, 6]
        assert len(data1) == len(data2)  # Valid same length

        # Length mismatch should be caught
        data3 = [1, 2]
        data4 = [3, 4, 5]
        assert len(data3) != len(data4)  # Invalid different length

    def test_schema_validation(self):
        """Test schema validation patterns"""
        fields_info = [
            {"name": "id", "is_primary": True, "auto_id": False},
            {"name": "vector", "is_primary": False}
        ]

        # Valid data matches schema
        data = {"id": [1, 2], "vector": [[1.0, 2.0], [3.0, 4.0]]}
        for field in fields_info:
            if not field.get("auto_id"):
                assert field["name"] in data  # Field should be in data

        # Test auto_id logic
        fields_info[0]["auto_id"] = True
        # When auto_id is True, id field should not be required in data






    def test_convert_to_entity(self):
        """Test converting field data to entities"""
        # Create field data
        field1 = schema_types.FieldData()
        field1.field_name = "id"
        field1.type = DataType.INT64
        long_array = schema_types.LongArray()
        long_array.data.extend([1, 2, 3])
        field1.scalars.long_data.CopyFrom(long_array)

        field2 = schema_types.FieldData()
        field2.field_name = "name"
        field2.type = DataType.VARCHAR
        str_array = schema_types.StringArray()
        str_array.data.extend(["a", "b", "c"])
        field2.scalars.string_data.CopyFrom(str_array)

        # extract_row_data_from_fields_data_v2 takes a single field_data and list of entity dicts
        # We need to create empty entity rows first
        entity_rows = [{} for _ in range(3)]

        # Process each field separately
        convert_to_entity(field1, entity_rows)
        convert_to_entity(field2, entity_rows)

        assert len(entity_rows) == 3
        assert entity_rows[0] == {"id": 1, "name": "a"}
        assert entity_rows[1] == {"id": 2, "name": "b"}
        assert entity_rows[2] == {"id": 3, "name": "c"}

    def test_entity_to_field_data(self):
        """Test converting entity to field data"""
        # entity_to_field_data expects a dict with specific structure
        entity = {
            "name": "test_field",
            "type": DataType.INT64,
            "values": [1, 2, 3, 4, 5]
        }
        field_info = {"name": "test_field"}

        result = entity_to_field_data(entity, field_info, 5)

        assert result.field_name == "test_field"
        assert result.type == DataType.INT64
        assert list(result.scalars.long_data.data) == [1, 2, 3, 4, 5]
